// FloodFill.cpp

#include "FloodFill.h"
#include "Voxel.h"
#include <cmath>

FloodFill::FloodFill() {}

FloodFill::~FloodFill() {}


bool FloodFill::colorsAreEqual(float col1, float col2)
{
  static const float margin = 0.01;
  float diff = fabs(col1 - col2);
  if (diff > margin) return false; // color of voxel is not equal to target color
  else return true;
}

void FloodFill::floodFillStack(short x, short y, short z, float& tColor, float& rColor)
{
  if (x > GRID_SIZE-1 || y > GRID_SIZE-1 || z > GRID_SIZE-1 || x < 0 || y < 0 || z < 0) return;
  if (voxels[x][y][z] == NULL) return;
  if (!colorsAreEqual(tColor, voxels[x][y][z]->getIntensity())) return;

  voxels[x][y][z]->setIntensity(rColor); // set color of voxel to target color
  
  floodFillStack(x+1, y, z, tColor, rColor);
  floodFillStack(x-1, y, z, tColor, rColor);
  floodFillStack(x, y+1, z, tColor, rColor);
  floodFillStack(x, y-1, z, tColor, rColor);
  floodFillStack(x, y, z+1, tColor, rColor);
  floodFillStack(x, y, z-1, tColor, rColor);
  return;
}

void FloodFill::floodFillQueueHelper(short x, short y, short z, vector<VoxIndex>& queue, float& tColor, float& rColor)
{
  if (x < GRID_SIZE && y < GRID_SIZE && z < GRID_SIZE && x >= 0 && y >= 0 && z >= 0) {
      if (voxels[x][y][z] == NULL) return;
      if (colorsAreEqual(tColor, voxels[x][y][z]->getIntensity())) {
        voxels[x][y][z]->setIntensity(rColor);
        struct VoxIndex vi = {x, y, z};
        queue.push_back(vi);
      }
    }
}

void FloodFill::floodFillQueue(short x, short y, short z, float tColor, float rColor)
{
  //ftime(&start_time); // function timing
  if (voxels[x][y][z] == NULL) return;
  vector<VoxIndex> queue;
  if (!colorsAreEqual(voxels[x][y][z]->getIntensity(), tColor)) return;
  struct VoxIndex voxInd = {x, y, z};
  queue.push_back(voxInd);
  
  while (!queue.empty()) {
    //display();
    struct VoxIndex n = queue.back();
    if (colorsAreEqual(voxels[n.x][n.y][n.z]->getIntensity(), tColor)) voxels[n.x][n.y][n.z]->setIntensity(rColor);
    queue.pop_back();
    
    floodFillQueueHelper(n.x+1, n.y, n.z, queue, tColor, rColor);
    floodFillQueueHelper(n.x-1, n.y, n.z, queue, tColor, rColor);
    floodFillQueueHelper(n.x, n.y+1, n.z, queue, tColor, rColor);
    floodFillQueueHelper(n.x, n.y-1, n.z, queue, tColor, rColor);
    floodFillQueueHelper(n.x, n.y, n.z+1, queue, tColor, rColor);
    floodFillQueueHelper(n.x, n.y, n.z-1, queue, tColor, rColor);
    
  }

  // calculate running time of function
  //ftime(&current_time);
  //ms_since_start = (long) (1000.0 * (current_time.time - start_time.time) + (current_time.millitm - start_time.millitm));
  //printf("ms since start: %d \n", ms_since_start);
}

VoxIndex FloodFill::findFirstValidVoxel()
{
  for (short i = 0; i < GRID_SIZE; ++i) {
    for (short j = 0; j < GRID_SIZE; ++j) {
      for (short k = 0; k < GRID_SIZE; ++k) {
        if (voxels[i][j][k]) {
          struct VoxIndex vi = {i, j, k};
          return vi;
        }
      }
    }
  }
}

// under construction (to be used by flood fill)
void FloodFill::placeSeed()
{
  struct VoxIndex vi = findFirstValidVoxel();
  float voxVal = voxels[vi.x][vi.y][vi.z]->getIntensity();
  float seedVal = voxVal+0.2f;
  if (seedVal > 0.8) seedVal = 0.0f;
}
