// Input.cpp

#include "Input.h"
#include "Globals.h"
#include "Voxel.h"
#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <iterator>

using std::cout;
using std::vector;
using std::endl;
using std::ifstream;
using std::ios;
using std::string;
using std::stringstream;

const char Input::delim_char = ' ';

Input::Input() {}

Input::~Input() {}

//process voxel slice
void Input::processVXS(short threshold)
{
  if (GRID_SIZE != 512) {
    cout << "Invalid grid size. Size must be 512 for VXS data.\n";
    exit(0);
  }
  short numFiles = 246; // max 246 (slices)
  string line;
  string sep("#");

  for (short curFile = 1; curFile <= numFiles; ++curFile) {
    string filename;
    stringstream out;
    out << "/Users/erlendlor/thesis/experiments/data/" << curFile << ".vxs";
    filename = out.str();
    ifstream inFile(filename.c_str());
    if (inFile.is_open()) {
      int y = 0;
      int x = 0;
      while (!inFile.eof()) {
        getline(inFile, line);
        if (line == sep) {
          //cout << "sep detected: " << line << endl;
          y++;
          x = 0;
        } else {
          //cout << "Size of x: " << x << endl;
          //cout << "Size of y: " << y << endl;
          //cout << atoi(line.c_str(short& x, short& y, short& z)) << endl;
          //slices[x][y][0] = atoi(line.c_str());
          short val = (short)atoi(line.c_str());
          if (val > threshold) voxels[x][y][curFile-1] = new Voxel((unsigned short)atoi(line.c_str()));
          x++;
        }
      }
      //cout << "Size of x: " << x << endl;
      //cout << "Size of y: " << y << endl;
      inFile.close();
    } else {
      cout << "Unable to open file\n";
    }
  }
}

void Input::processPPM(string filename)
{
  string line;
  int xSize = 0;
  int ySize = 0;
  ifstream inFile(filename.c_str());

  if (inFile.is_open()) {
    getline(inFile, line);
    vector<string> elems;
    elems = split(line, delim_char, elems);
    if (elems.size() != 2) { cout << "Invalid file."; exit(0); }
    xSize = atoi(elems[0].c_str());
    ySize = atoi(elems[1].c_str());
    
    cout << "xSize: " << xSize << endl;
    cout << "ySize: " << ySize << endl;

    getline(inFile, line); // skip max value
    cout << "max value: " << line << endl;

    for (int y = ySize-1; y >= 0; --y) {
      if (inFile.eof()) break;
      for (int x = 0; x < xSize; ++x) {
        getline(inFile, line);
        if (inFile.eof()) break;
        short R = (short)atoi(line.c_str());
        getline(inFile, line);
        if (inFile.eof()) break;
        short G = (short)atoi(line.c_str());
        getline(inFile, line);
        if (inFile.eof()) break;
        short B = (short)atoi(line.c_str());
        unsigned short voxval = (R+G+B)/3;
        
        voxels[x][y][GRID_SIZE/2] = new Voxel(voxval);

        //cout << "x: " << x << endl;
        //cout << "y: " << y << endl;
        //cout << "voxval: " << voxval << endl;
      }
    }
    inFile.close();
  } else {
    cout << "Unable to open file\n";
  }
}

vector<string>& Input::split(const string& s, const char delim, vector<string>& elems) {
  stringstream ss(s);
  string item;
  while(getline(ss, item, delim)) {
    elems.push_back(item);
  }
  return elems;
}
