// Main.cpp
//

#include "Globals.h"
#include "Main.h"

// prototypes
void benchmarkS3DRG(short neighborhood);
void testS3DRG_FIFO_LIFO();
void benchmarkSRG3D();
void benchmarkASRG3D(short neighborhood);


void reshape(int width, int height)
{
  float widthF = (float) width;
  float heightF = (float) height;
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  //glOrtho(-30, 30, -30, 30, -30, 30);
  gluPerspective(90, widthF/heightF, 0.1, 200.0);
  glViewport(0, 0, width, height);
  glMatrixMode(GL_MODELVIEW);
}


void toggleShading()
{
  if (flatShading) flatShading = false;
  else flatShading = true;
}

void toggleLighting()
{
  if (lighting) lighting = false;
  else lighting = true;
}

void toggleDrawVoxels()
{
  if (drawVoxelsEnabled) drawVoxelsEnabled = false;
  else drawVoxelsEnabled = true;
}

void toggleDrawMesh()
{
  if (drawMeshEnabled) drawMeshEnabled = false;
  else drawMeshEnabled = true;
}

void drawVoxelsOldBenchmark()
{
  ftime(&start_time); // function timing
  int numVoxels = 0;
  if (drawVoxelsEnabled) {
    for (short i = 0; i < GRID_SIZE; ++i) {
      for (short j = 0; j < GRID_SIZE; ++j) {
        for (short k = 0; k < GRID_SIZE; ++k) {
          numVoxels++;
          if (voxels[i][j][k]) {
            if (i == 0 || j == 0 || k == 0 || i == GRID_SIZE-1 || j == GRID_SIZE-1 || k == GRID_SIZE-1) { // always render border voxels
              //renderVoxels.push_back(voxels[i][j][k]); // store pointer to voxel
              //renderVoxelsIndex.push_back(createVoxelIndex(i,j,k)); // store voxel index
              voxels[i][j][k]->draw(i,j,k);
            } 
            else {
              if (voxels[i+1][j][k] && voxels[i-1][j][k] &&
                  voxels[i][j+1][k] && voxels[i][j-1][k] &&
                  voxels[i][j][k+1] && voxels[i][j][k-1]) {
                // not visible, do not render
              } else { // render visible voxel
                //renderVoxels.push_back(voxels[i][j][k]); // store pointer to voxel
                //renderVoxelsIndex.push_back(createVoxelIndex(i,j,k)); // store voxel index
                voxels[i][j][k]->draw(i,j,k);
              }
            }
          }
        }
      }
    }
  }
  // calculate running time of function
  ftime(&current_time);
  ms_since_start = (long) (1000.0 * (current_time.time - start_time.time) + (current_time.millitm - start_time.millitm));
  printf("ms since start (old render method): %d \n", ms_since_start);
  cout << "Number of voxels: " << numVoxels << endl;
  cout << "Number of voxels to render: " << renderVoxels.size() << endl;
}

void drawVoxelsNewBenchmark()
{
  ftime(&start_time); // function timing
  if (drawVoxelsEnabled) {
    for (int i = 0; i < renderVoxels.size(); ++i) {
      if (!drawVoxelsEnabled) break;
      if (renderVoxels.at(i)) renderVoxels.at(i)->draw(renderVoxelsIndex.at(i).at(0),
                                                       renderVoxelsIndex.at(i).at(1),
                                                       renderVoxelsIndex.at(i).at(2));
    }
  }
  // calculate running time of function
  ftime(&current_time);
  ms_since_start = (long) (1000.0 * (current_time.time - start_time.time) + (current_time.millitm - start_time.millitm));
  printf("ms since start (new render method): %d \n", ms_since_start);
}

void drawVoxelsOld()
{
  int numVoxels = 0;
  if (drawVoxelsEnabled) {
    for (short i = 0; i < GRID_SIZE; ++i) {
      for (short j = 0; j < GRID_SIZE; ++j) {
        for (short k = 0; k < GRID_SIZE; ++k) {
          numVoxels++;
          if (voxels[i][j][k]) {
            if (i == 0 || j == 0 || k == 0 || i == GRID_SIZE-1 || j == GRID_SIZE-1 || k == GRID_SIZE-1) { // always render border voxels
              //renderVoxels.push_back(voxels[i][j][k]); // store pointer to voxel
              //renderVoxelsIndex.push_back(createVoxelIndex(i,j,k)); // store voxel index
              if (voxels[i][j][k]->getLabel() == renderLabel) voxels[i][j][k]->draw(i,j,k);
            } 
            else {
              if (voxels[i+1][j][k] && voxels[i-1][j][k] &&
                  voxels[i][j+1][k] && voxels[i][j-1][k] &&
                  voxels[i][j][k+1] && voxels[i][j][k-1]) {
                // not visible, do not render
                if (voxels[i][j][k]->getLabel() == renderLabel) voxels[i][j][k]->draw(i,j,k);
              } else { // render visible voxel
                //renderVoxels.push_back(voxels[i][j][k]); // store pointer to voxel
                //renderVoxelsIndex.push_back(createVoxelIndex(i,j,k)); // store voxel index
                if (voxels[i][j][k]->getLabel() == renderLabel) voxels[i][j][k]->draw(i,j,k);
              }
            }
          }
        }
      }
    }
  }
}

void drawVoxels()
{
  if (drawVoxelsEnabled) {
    for (int i = 0; i < renderVoxels.size(); ++i) {
      if (!drawVoxelsEnabled) break;
      if (renderVoxels.at(i)) {
        //if (renderVoxels.at(i)->getLabel() == renderLabel) {
          renderVoxels.at(i)->draw(renderVoxelsIndex.at(i).at(0),
                                   renderVoxelsIndex.at(i).at(1),
                                   renderVoxelsIndex.at(i).at(2));
        //}
      }
    }
  }
}

void drawMesh()
{
  if (drawMeshEnabled) {
    vector<Triangle*>::iterator it;
    for (it = mesh.begin(); it < mesh.end(); ++it) {
      (*it)->draw();
    }
  }
}

void setShading(char t)
{
  if (t == 'f') glShadeModel(GL_FLAT);
  else if (t == 's') glShadeModel(GL_SMOOTH);
}

void enableLighting()
{
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
}

void disableLighting()
{
  glDisable(GL_LIGHT1);
  glDisable(GL_LIGHT0);
  glDisable(GL_LIGHTING);
}

void display()
{
  // setup
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(cam.eyeX, cam.eyeY, cam.eyeZ, cam.atX, cam.atY, cam.atZ, cam.upX, cam.upY, cam.upZ);

  // transform
  glTranslatef(transFactor, transFactor, -transFactor);
  glRotatef(animeTheta, 0.0, 1.0, 0.0);
  glTranslatef(-transFactor, -transFactor, transFactor);

  // draw
  drawVoxels();
  drawMesh();
  
  // toggle options
  if (lighting) enableLighting();
  else disableLighting();
  if (flatShading) setShading('f');
  else setShading('s');

  glutSwapBuffers();
}

void rotateObj()
{
  if (rotate) {
    t.tv_sec = 0.0;
    t.tv_nsec = 10000000;
    nanosleep(&t, 0);
    animeTheta += 2.0f;
    if (animeTheta > 360.0f) animeTheta -= 360.0f;
  }
}

void idle()
{
  rotateObj();
  glutPostRedisplay();
}

void keyFunc(unsigned char key, int x, int y)
{
  if (key == 'k') {
    renderLabel++;
  }
  if (key == 'q') exit(0);
  if (key == 'r') if (rotate) rotate = false; else rotate = true;
  if (key == 'w') {
    Voxel::toggleWireframe();
    Triangle::toggleWireframe();
  }
  if (key == 'f') /*benchmarkS3DRG(6);*/ benchmarkASRG3D(6); /* testS3DRG_FIFO_LIFO();*/ /*benchmarkSRG3D();*/
  if (key == 'x') Models::deleteVoxel(0,0,0);
  if (key == 'm') MarchingCubes::marchingCubes();
  if (key == 'd') toggleDrawMesh();
  if (key == 'v') toggleDrawVoxels();
  if (key == 'l') toggleLighting();
  if (key == 'u') {
    Models::updateRenderVoxels();
    Segmentation::currentlySelectedSegment = -1;
  }
  if (key == 's') {
    if (Segmentation::getSegmentSurfaces().size() == 0) return;
    Segmentation::currentlySelectedSegment++;
    unsigned short numberOfSegments = Segmentation::getSegmentSurfaces().size();
    if (Segmentation::currentlySelectedSegment >= numberOfSegments) Segmentation::currentlySelectedSegment = 0;
    Models::setRenderVoxels(Segmentation::getSegmentSurfaces().at(Segmentation::currentlySelectedSegment));
    cout << "Currently selected segment: " << Segmentation::currentlySelectedSegment+1 << endl;
  }
  if (key == 'h') {
    if (Segmentation::getSegments().size() == 0) return;
    if (Segmentation::renderSegmentSurface) {
      Models::setRenderVoxels(Segmentation::getSegments().at(Segmentation::currentlySelectedSegment));
      Segmentation::renderSegmentSurface = false;
    } else {
      Models::setRenderVoxels(Segmentation::getSegmentSurfaces().at(Segmentation::currentlySelectedSegment));
      Segmentation::renderSegmentSurface = true;
    }
  }
  if (key == 'b') Models::booleanSubtract();
  if (key == 'p') Voxel::togglePseudoColor();
  if (key == 'c') Voxel::togglePoint();
  if (key == 'a') toggleShading();
  if (key == '1') {
    cam.eyeX = transFactor;
    cam.eyeY = transFactor;
    cam.eyeZ = transFactor;
  }
  if (key == '2') {
    cam.eyeX = 0.0;;
    //cam.eyeY = transFactor;
    cam.eyeZ = transFactor / 10.0;
  }
  if (key == '3') {
    cam.eyeX = transFactor / 10.0;
    //cam.eyeY = transFactor;
    cam.eyeZ = transFactor / 10.0;
  }
  if (key == '4') {
   cam.eyeY -= (transFactor/20.0);
   cam.eyeZ += (transFactor/30.0);
  }
  if (key == '5') {
    cam.eyeY += (transFactor/20.0);
    cam.eyeZ -= (transFactor/30.0);
  }
}

void specialFunc(int key, int x, int y)
{
  float stepSize = 0.1; // 0.6
  if (key == GLUT_KEY_UP) cam.eyeZ = cam.eyeZ-stepSize;
  if (key == GLUT_KEY_DOWN) cam.eyeZ = cam.eyeZ+stepSize;
  if (key == GLUT_KEY_LEFT) cam.eyeX = cam.eyeX-stepSize;
  if (key == GLUT_KEY_RIGHT) cam.eyeX = cam.eyeX+stepSize;
}

void initLighting()
{
  GLfloat light_pos0[] = {cam.eyeX+3.0, cam.eyeY+6.0, cam.eyeZ-1.2, 1.0};
  GLfloat light_pos1[] = {cam.eyeX+0.5, cam.eyeY-0.2, cam.eyeZ-3.0, 1.0};
  GLfloat ambient[] = {.2, .2, .2, 1.0};
  GLfloat diffuse[] = {0.9, 0.9, 0.9, 1.0};
  GLfloat specular[] = {0.9, 0.9, 0.9, 1.0};

  glLightfv(GL_LIGHT0, GL_POSITION, light_pos0);
  glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, specular);
  
  glLightfv(GL_LIGHT1, GL_POSITION, light_pos1);
  glLightfv(GL_LIGHT1, GL_AMBIENT, ambient);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse);
  glLightfv(GL_LIGHT1, GL_SPECULAR, specular);
  
  GLfloat matAmbient[4] = {0.05, 0.05, 0.05, 1.0}; // 0.05
  GLfloat matDiffuse[4] = {0.6, 0.6, 0.6, 1.0}; // 0.3
  GLfloat matSpecular[4] = {0.5, 0.5, 0.5, 1.0}; //0.5
  GLfloat matEmission[4] = {0.05, 0.05, 0.05, 1.0}; //0.05
  GLfloat matShininess = 50.0; // 50.0
  Material material(matAmbient, matDiffuse, matSpecular, matEmission, matShininess);

  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material.ambient);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material.diffuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material.specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, material.emission);
  glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, material.shininess);
   
}

void initOpenGL(int argc, char ** argv)
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  glutInitWindowSize(800, 600);
  glutCreateWindow("VoxVis");
  glutKeyboardFunc(keyFunc);
  glutSpecialFunc(specialFunc);
  glutReshapeFunc(reshape);
  glutIdleFunc(idle);
  glutDisplayFunc(display);

  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  glPointSize(15.0);
  glEnable(GL_DEPTH_TEST);
  glShadeModel(GL_FLAT);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_LINE_SMOOTH);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  //glEnable(GL_CULL_FACE);
  //glCullFace(GL_BACK);
  glClearColor(0.0, 0.0, 0.0, 1.0);

  initLighting();
}

void initTrans()
{
  transFactor = (GRID_SIZE/2.0)*(Voxel::getDim());
  cout << "GRID_SIZE: " << GRID_SIZE << endl;
  cout << "dim: " << Voxel::getDim() << endl;
  cout << "transFactor: " << transFactor << endl;
  cam.eyeX = transFactor;
  cam.eyeY = transFactor;
  cam.eyeZ = transFactor;
  cam.atX = transFactor;
  cam.atY = transFactor;
  cam.atZ = -transFactor;
}

void debug()
{
  cout << "int: " << sizeof(int) << endl;
  cout << "unsigned int: " << sizeof(unsigned int) << endl;
  cout << "float: " << sizeof(float) << endl;
  cout << "short: " << sizeof(short) << endl;
  cout << "Voxel: " << sizeof(Voxel) << endl;
  cout << "bool: " << sizeof(bool) << endl;
}


vector<short> findFirstValidVoxel()
{
  vector<short> vi;
  for (short i = 0; i < GRID_SIZE; ++i) {
    for (short j = 0; j < GRID_SIZE; ++j) {
      for (short k = 0; k < GRID_SIZE; ++k) {
        if (voxels[i][j][k]) {
          vi.push_back(i);
          vi.push_back(j);
          vi.push_back(k);
          return vi;
        }
      }
    }
  }
  return vi;
}

void benchmarkS2DRG(short neighborhood)
{
  ftime(&start_time); // function timing
  Segmentation::growRegionsS2DRG(neighborhood);
  // calculate running time of function
  ftime(&current_time);
  ms_since_start = (long) (1000.0 * (current_time.time - start_time.time) + (current_time.millitm - start_time.millitm));
  printf("S2DRG running time: %d ms\n", ms_since_start);
}

void benchmarkS3DRG(short neighborhood)
{
  ftime(&start_time); // function timing
  cout << "Starting segmentation...\n";
  Segmentation::growRegionsS3DRG(display, neighborhood);
  // calculate running time of function
  ftime(&current_time);
  ms_since_start = (long) (1000.0 * (current_time.time - start_time.time) + (current_time.millitm - start_time.millitm));
  printf("S3DRG running time: %d ms\n", ms_since_start);
}

void benchmarkSRG3D()
{
  ftime(&start_time); // function timing
  cout << "Startig segmentation...\n";
  
  struct VoxIndex seed1 = {0, 0, 0};
  struct VoxIndex seed2 = {63, 63, 0};
  struct VoxIndex seed3 = {63, 0, 0};
  struct VoxIndex seed4 = {60, 0, 0};
  struct VoxIndex seed5 = {55, 0, 0};
  
  struct VoxIndex seed6 = {56, 62, 0};
  struct VoxIndex seed7 = {56, 59, 0};
  struct VoxIndex seed8 = {56, 55, 0};
  vector<struct VoxIndex> seeds;
  seeds.push_back(seed1);
  seeds.push_back(seed2);
  seeds.push_back(seed3);
  seeds.push_back(seed4);
  seeds.push_back(seed5);
  seeds.push_back(seed6);
  seeds.push_back(seed7);
  seeds.push_back(seed8);
  bool heuristicSearchOn = true;
  Segmentation::SRG3D(seeds, heuristicSearchOn);
  // calculate running time of function
  ftime(&current_time);
  ms_since_start = (long) (1000.0 * (current_time.time - start_time.time) + (current_time.millitm - start_time.millitm));
  printf("SRG3D running time: %d ms\n", ms_since_start);
}

void benchmarkASRG3D(short neighborhood)
{
  ftime(&start_time);
  cout << "Starting segmentation...\n";
  Segmentation::growRegionsS3DRG(display, neighborhood);
  vector<struct VoxIndex> seeds = Segmentation::findSeeds();
  bool heuristicSearchOn = true;
  Segmentation::SRG3D(seeds, heuristicSearchOn);
  ftime(&current_time);
  ms_since_start = (long) (1000.0 * (current_time.time - start_time.time) + (current_time.millitm - start_time.millitm));
  printf("ASRG3D running time: %d ms\n", ms_since_start);
}

void testS3DRG_FIFO_LIFO()
{
  Segmentation::floodFillRegionS3DRG(40, 40, 0, 0.77f, display);
}

void init(int argc, char ** argv)
{
  initOpenGL(argc, argv);
  Models::initVoxelGrid();
  initTrans();
  debug();
  
  short threshold = 1500; //200 // 1500
  Input::processVXS(threshold);
  //string filename = "/Users/erlendlor/thesis/experiments/data/ppm/ling.ppm";
  //Input::processPPM(filename);
  //processInputFiles();

  //Models::plotTestVoxels();
  //Models::plot2DFunc();
  //Models::plot3DFunc();
  //Models::plot3DGasket();
  //Models::plotSphere();
  //Models::plotCube();
  
  //Models::plotBorder();
  //Models::plotModelGroup1();
  //Models::plotModelGroup2();
  //Models::plotSimpleModel1();
  
  //Models::plotSimpleModel2(); // grid size 4
  //Models::plotSegTest1(); // grid size 0 -> 512
  //Models::plotBiasTest(); // grid size 6
  //Models::plotFIFOtest();
  //Models::plotSRGTest1(); // grid size 6
  //Models::plotSRGTest2(); // grid size 0 -> 512
  
  //vector<short> vi = findFirstValidVoxel();
  //cout << "Label value #####: " << voxels[vi[0]][vi[1]][vi[2]]->getLabel() << endl;
  //cout << "Intensity value #####: " << voxels[vi[0]][vi[1]][vi[2]]->getIntensityShort() << endl;
  
  // segmentation
  //benchmarkASRG3D(26);
  //benchmarkS2DRG(8);
  //benchmarkS3DRG(26);
  //benchmarkSRG3D();
  //Segmentation::growRegions();
  //cout << "Label value #####: " << voxels[vi[0]][vi[1]][vi[2]]->getLabel() << endl;
  //cout << "Intensity value #####: " << voxels[vi[0]][vi[1]][vi[2]]->getIntensityShort() << endl;
  
  
  //Models::plotTestVoxels();
  //Models::plotSimpleModel();
  //Models::plotVoxel(GRID_SIZE-3, 0, 0, 0.5);
  //Models::plotGrid();
  //Models::plotGridWithBorder();
  //Models::plotBorg();
  //FloodFill::floodFillStack(GRID_SIZE-3, 0, 0, 0.9, 0.5);

  //secondary plots
  //Models::initSecVoxelGrid();
  //Models::plotSecSphere();
  
  //build render vector
  Models::updateRenderVoxels();
  //drawVoxelsOldBenchmark();
  //drawVoxelsNewBenchmark();

  glutMainLoop();
  Models::deleteAllVoxels();
}

int main(int argc, char ** argv)
{
  srand(time(NULL)); // initialize rand with seed
  init(argc, argv);
  return 0;
}
