// Main.h

#ifndef MAIN_H_
#define MAIN_H_

#include "Voxel.h"
#include "VoxIndex.h"
#include "Triangle.h"
#include "Material.h"
#include "MarchingCubes.h"
#include "FloodFill.h"
#include "Models.h"
#include "Input.h"
#include "Segmentation.h"
#include <glut/glut.h>
#include <vector>
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <ctime>
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <sys/timeb.h>

using std::cout;
using std::vector;
using std::endl;
using std::ifstream;
using std::ios;
using std::string;
using std::stringstream;

// OpenGL globals
bool lighting = true;
bool flatShading = false;

// Voxel globals
bool drawVoxelsEnabled = true;
short renderLabel = 1;
short GRID_SIZE = 512;
int NUM_VOXELS = GRID_SIZE * GRID_SIZE * GRID_SIZE;
vector<vector<vector<Voxel*> > > voxels;
short SEC_GRID_SIZE = GRID_SIZE;
int SEC_NUM_VOXELS = SEC_GRID_SIZE * SEC_GRID_SIZE * SEC_GRID_SIZE;
vector<vector<vector<Voxel*> > > sec_voxels;
bool secVoxelsExist = false;
vector<Voxel*> renderVoxels;
vector<vector<short> > renderVoxelsIndex;
const unsigned short UNLABELED = 65535;
const unsigned short NOISE = 65534;
const unsigned short SSL_ELEM = 65533;

// Marching Cube globals
bool drawMeshEnabled = true;
vector<Triangle*> mesh;

// Timing and animation globals
struct timespec t;
bool rotate = false;
float animeTheta = 0.0;
struct timeb start_time, current_time;
long ms_since_start;

// Translation and camera globals
float transFactor = 0.0;
struct Camera
{
  float eyeX, eyeY, eyeZ;
  float atX, atY, atZ;
  float upX, upY, upZ;
} cam = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0 };

/*
struct Vec
{
  float x; float y; float z;
};
*/
 
#endif // MAIN_H_
