// Models.cpp

#include "Models.h"
#include "Globals.h"
#include "Voxel.h"
#include "Triangle.h"
#include "Material.h"
#include "MarchingCubes.h"
#include "FloodFill.h"
#include "Models.h"
#include <glut/glut.h>
#include <vector>
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <ctime>
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <sys/timeb.h>

Models::Models() {}

Models::~Models() {}

void Models::plotTestVoxels()
{
  int shape = 0;
  float color = 0.01;
  for (short i = 0; i < GRID_SIZE; ++i) {
    for (short j = 0; j < GRID_SIZE; ++j) {
      for (short k = 0; k < GRID_SIZE; ++k) {
        int randNum = rand() % 100 + 1;
        int randCol = rand() % 95 + 1; // 50 + 1
        if (randNum < 101) {
          color = randCol * 0.01;
          if (color <= 0.0 || color > 1.0) color = 0.8;
          if (color > 0.0) color = 0.8;
          if (shape == 0) {
            if (i < GRID_SIZE/2) {plotVoxel(i, j, k, color);}
            else {plotVoxel(i, j, k, 0.5f);}
            float noiseColor = 0.1f;
            plotVoxel(7, 4, 0, noiseColor);
            plotVoxel(29, 14, 0, noiseColor);
            plotVoxel(0, 8, 20, noiseColor);
            plotVoxel(12, 31, 9, noiseColor);
          }
          if (shape == 1 &&
              (i == j || i == k || j == k)) {
            plotVoxel(i, j, k, 0.3f);
          }
          if (shape == 2 &&
              (i == (short)pow(j, 2.0) ||
               i == (short)pow(k, 2.0) ||
               j == (short)pow(k, 2.0))) {
            plotVoxel(i, j, k, color);
          }
          if (shape == 3) {
            plotVoxel(i, j, k, color);
          }
        }
      }
    }
  }
}

void Models::plotNoise(int factor)
{
  float color = 0.01;
  for (short i = 0; i < GRID_SIZE; ++i) {
    for (short j = 0; j < GRID_SIZE; ++j) {
      for (short k = 0; k < GRID_SIZE; ++k) {
        int randNum = rand() % 10000 + 1;
        int randCol = rand() % 50 + 1;
        if (randNum < factor) {
          color = randCol * 0.01;
          if (color <= 0.0 || color > 1.0) color = 1.0;
          plotVoxel(i, j, k, color);
        }
      }
    }
  }
}

void Models::plot3DFunc()
{
  for (short x = 1; x < GRID_SIZE-1; ++x) {
    for (short y = 1; y < GRID_SIZE-1; ++y) {
      for (short z = 1; z < GRID_SIZE-1; ++z) {
        if (pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) >= (double)GRID_SIZE*GRID_SIZE/1.07 &&
            pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) <= (double)GRID_SIZE*GRID_SIZE) {
          int randNum = rand() % 40 +  35;
          float color = randNum * 0.01;
           
          plotVoxel(x, y, z, 0.1f); // color, temp val
        }
        if (pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) >= (double)GRID_SIZE*GRID_SIZE/1.3 &&
            pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) <= (double)GRID_SIZE*GRID_SIZE/1.07) {
          int randNum = rand() % 40 +  35;
          float color = randNum * 0.01;
          plotVoxel(x, y, z, 0.9f);
        } 
        if (pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) >= (double)GRID_SIZE*GRID_SIZE/1.5 &&
            pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) <= (double)GRID_SIZE*GRID_SIZE/1.3) {
          int randNum = rand() % 40 +  35;
          float color = randNum * 0.01;
          plotVoxel(x, y, z, 0.1f);
        } 
      }
    }
  }
}

/*
void Models::plotSimpleModel()
{
  voxels[(GRID_SIZE/2)+0][(GRID_SIZE/2)+0][(GRID_SIZE/2)+1] = new Voxel(0.9f);
  voxels[(GRID_SIZE/2)+1][(GRID_SIZE/2)+0][(GRID_SIZE/2)+1] = new Voxel(0.9f);
  voxels[(GRID_SIZE/2)+2][(GRID_SIZE/2)+0][(GRID_SIZE/2)+1] = new Voxel(0.9f);
  voxels[(GRID_SIZE/2)+1][(GRID_SIZE/2)+1][(GRID_SIZE/2)+1] = new Voxel(0.9f);
  voxels[(GRID_SIZE/2)+1][(GRID_SIZE/2)+0][(GRID_SIZE/2)+0] = new Voxel(0.9f);
}
*/

void Models::plot2DFunc()
{
  for (short x = 0; x < GRID_SIZE; ++x) {
    double sinScaled = sin(x)*(GRID_SIZE/2);
    short y = short(sinScaled + (GRID_SIZE/2));
    for (short z = 0; z < GRID_SIZE; ++z) {
      int randNum = rand() % 100 + 1;
      float color = randNum * 0.02;
      plotVoxel(x, y, z, color);
    }
    //cout << "sin x: " << sinTrans << endl;
  }
}


// Creates a sphere
// Reference model: rotation == 0, scale == 0, transaltion == 0, noise == 0
// Scale range: 0.0 -> 0.5
// Noise range: 0 -> 10000 (0 -> 100 is a more sane range)
// Translation range: -GRID_SIZE -> GRID_SIZE
void Models::plotSphere(float intensity, double scale, short xTrans, short yTrans, short zTrans, int noise)
{
  short c = GRID_SIZE/2;
  if (scale <= 0.0) scale = 0.05;
  if (scale > 0.48) scale = 0.48;
  
  for (short x = 0; x < GRID_SIZE; ++x) {
    for (short y = 0; y < GRID_SIZE; ++y) {
      for (short z = 0; z < GRID_SIZE; ++z) {
        if (pow(x-c, 2.0) + pow(y-c, 2.0) + pow(z-c, 2.0) <= (double)pow(GRID_SIZE/(1.0/scale), 2.0)) {
          int randNum = rand() % 40 + 35;
          float color = randNum * 0.01;
          //cout << x+GRID_SIZE << ", " << y+GRID_SIZE << ", " << z+GRID_SIZE << endl;
          plotVoxel(x+xTrans, y+yTrans, z+zTrans, intensity); // color, temp val
          //voxels[x+xTrans][y+yTrans][z+zTrans]->setLabel(2);
        }
      }
    }
  }
  plotNoise(noise);
}

// Creates a cube
// Scale range: GRID_SIZE/2 -> 0 (0 is the largest cube)
// Translation range: -GRID_SIZE -> GRID_SIZE
void Models::plotCube(float intensity, short scale, short xTrans, short yTrans, short zTrans, int rotation, int noise)
{
  
  for (int i = scale; i < GRID_SIZE-scale; ++i) {
    for (int j = scale; j < GRID_SIZE-scale; ++j) {
      for (int k = scale; k < GRID_SIZE-scale; ++k) {
        plotVoxel(i+xTrans, j+yTrans, k+zTrans, intensity);
        //voxels[i+xTrans][j+yTrans][k+zTrans]->setLabel(1);
      }
    }
  }
  plotNoise(noise);
}

void Models::plotBorder()
{
  for (int i = 0; i < GRID_SIZE; ++i) {
    for (int j = 0; j < GRID_SIZE; ++j) {
      for (int k = 0; k < GRID_SIZE; ++k) {
        if (i == 0 && j == 0 ||
            i == 0 && k == 0 ||
            j == 0 && k == 0 ||
            i == GRID_SIZE-1 && j == GRID_SIZE-1 ||
            i == GRID_SIZE-1 && k == GRID_SIZE-1 ||
            j == GRID_SIZE-1 && k == GRID_SIZE-1 ||
            i == 0 && j == GRID_SIZE-1 ||
            i == 0 && k == GRID_SIZE-1 ||
            j == 0 && k == GRID_SIZE-1 ||
            j == 0 && i == GRID_SIZE-1 ||
            k == 0 && i == GRID_SIZE-1 ||
            k == 0 && j == GRID_SIZE-1) {
          plotVoxel(i,j,k,0.9f);
        }
      }
    }
  }
}

void Models::plotSimpleModel1()
{
  // slice 1
  plotVoxel(0, 0, 0, 0.9);
  plotVoxel(1, 0, 0, 0.9);
  plotVoxel(2, 0, 0, 0.9);
  plotVoxel(0, 1, 0, 0.9);
  plotVoxel(1, 1, 0, 0.9);
  plotVoxel(2, 1, 0, 0.9);
  plotVoxel(0, 2, 0, 0.9);
  plotVoxel(1, 2, 0, 0.9);
  plotVoxel(2, 2, 0, 0.9);
  
  // slice 2
  plotVoxel(0, 0, 1, 0.9);
  plotVoxel(1, 0, 1, 0.9);
  plotVoxel(2, 0, 1, 0.9);
  plotVoxel(0, 1, 1, 0.9);
  plotVoxel(1, 1, 1, 0.9);
  plotVoxel(2, 1, 1, 0.9);
  plotVoxel(0, 2, 1, 0.9);
  plotVoxel(1, 2, 1, 0.9);
  plotVoxel(2, 2, 1, 0.9);
  
  // slice 3
  plotVoxel(0, 0, 2, 0.9);
  plotVoxel(1, 0, 2, 0.9);
  plotVoxel(2, 0, 2, 0.9);
  plotVoxel(0, 1, 2, 0.9);
  plotVoxel(1, 1, 2, 0.9);
  plotVoxel(2, 1, 2, 0.9);
  plotVoxel(0, 2, 2, 0.9);
  plotVoxel(1, 2, 2, 0.9);
  plotVoxel(2, 2, 2, 0.9);
  
  plotVoxel(3, 1, 1, 0.9);
  plotVoxel(4, 1, 1, 0.9);
  plotVoxel(4, 1, 0, 0.9);
  
  // new region
  plotVoxel(3, 0, 0, 0.3);
  plotVoxel(3, 1, 0, 0.3);
  plotVoxel(3, 2, 0, 0.3);
  plotVoxel(4, 3, 0, 0.3);
  plotVoxel(3, 0, 1, 0.3);
  plotVoxel(3, 1, 2, 0.3);
  plotVoxel(3, 2, 1, 0.3);
  plotVoxel(3, 0, 2, 0.3);
  plotVoxel(3, 2, 2, 0.3);
  
  // new region
  plotVoxel(0, 3, 0, 0.5);
  plotVoxel(1, 3, 0, 0.5);
  plotVoxel(1, 4, 1, 0.5);
  plotVoxel(2, 5, 2, 0.5);
  plotVoxel(3, 6, 3, 0.5);
}

void Models::plotSimpleModel2()
{
  plotVoxel(0, 0, 0, 0.4);
  plotVoxel(0, 1, 0, 0.4);
  plotVoxel(1, 2, 0, 0.4);
  plotVoxel(1, 3, 0, 0.4);
  plotVoxel(2, 0, 0, 0.4);
  plotVoxel(2, 1, 0, 0.4);
  plotVoxel(3, 2, 1, 0.4);
  floodFillNull(0.8);
}

void Models::plotModelGroup1()
{
  Models::plotSphere(0.6, 0.16, 20, 2, -4, 0);
  Models::plotSphere(0.6, 0.1, 22, 20, 11, 0);
  Models::plotSphere(0.15, 0.14, -20, 20, -25, 0);
  Models::plotSphere(0.45, 0.111, -10, 2, 29, 0);
  Models::plotSphere(0.3, 0.1, 23, -20, -20, 0);
  Models::plotSphere(0.85, 0.11, 20, -20, 20, 0);
  Models::plotSphere(0.9, 0.17, -14, -21, 26, 0);
  Models::plotSphere(0.95, 0.1, -20, -20, -2, 0);
  Models::plotCube(1.0, 34, 10, -14, 10, 0, 0);
  Models::plotCube(0.5, 36, 26, 27, 25, 0, 10);
}

void Models::plotModelGroup2()
{
  Models::plotCube(0.2, GRID_SIZE/2.5, -GRID_SIZE/5, GRID_SIZE/10, GRID_SIZE/10, 0, 0);
  Models::plotSphere(0.9, 0.15, GRID_SIZE/5, GRID_SIZE/10, GRID_SIZE/10, 0);
  
}

void Models::plotSegTest1()
{
  short c = GRID_SIZE/2;
  
  for (short x = 0; x < GRID_SIZE; ++x) {
    for (short y = 0; y < GRID_SIZE; ++y) {
      for (short z = 0; z < GRID_SIZE; ++z) {
        if (pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) >= (double)GRID_SIZE*GRID_SIZE/1.07 &&
            pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) <= (double)GRID_SIZE*GRID_SIZE) {
          plotVoxel(x, y, z, 0.17f); // color, temp val
        }
        if (pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) >= (double)GRID_SIZE*GRID_SIZE/1.3 &&
            pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) <= (double)GRID_SIZE*GRID_SIZE/1.07) {
          plotVoxel(x, y, z, 0.25f);
        } 
        if (pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) >= (double)GRID_SIZE*GRID_SIZE/1.5 &&
            pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) <= (double)GRID_SIZE*GRID_SIZE/1.3) {
          plotVoxel(x, y, z, 0.30f);
        }
        
        if (z < GRID_SIZE/4) {
          if (pow(x-c, 2.0) + pow(y-c, 2.0) <= (double)pow(GRID_SIZE/6.5, 2.0)) {
            plotVoxel(x+(GRID_SIZE/3), y+(GRID_SIZE/3), z, 0.9f);
          }
          if (pow(x-c, 2.0) + pow(y-c, 2.0) <= (double)pow(GRID_SIZE/8.0, 2.0)) {
            plotVoxel(x+(GRID_SIZE/3), y+(GRID_SIZE/3), z, 0.81f);
          }
          if (pow(x-c, 2.0) + pow(y-c, 2.0) <= (double)pow(GRID_SIZE/11.0, 2.0)) {
            plotVoxel(x+(GRID_SIZE/3), y+(GRID_SIZE/3), z, 0.77f);
          }
        }
      }
    }
  }
  floodFillNull(0.1);
  
  /* IMPORTANT: seed positions for 32^3 version of model
   struct VoxIndex seed1 = {0, 0, 0};
   struct VoxIndex seed2 = {31, 31, 0};
   struct VoxIndex seed3 = {31, 0, 0};
   struct VoxIndex seed4 = {29, 0, 0};
   struct VoxIndex seed5 = {27, 0, 0};
   
   struct VoxIndex seed6 = {28, 30, 0};
   struct VoxIndex seed7 = {26, 30, 0};
   struct VoxIndex seed8 = {26, 26, 0};
  */
  
  /* IMPORTANT: seed positions for 64^3 version of model
   struct VoxIndex seed1 = {0, 0, 0};
   struct VoxIndex seed2 = {63, 63, 0};
   struct VoxIndex seed3 = {63, 0, 0};
   struct VoxIndex seed4 = {60, 0, 0};
   struct VoxIndex seed5 = {55, 0, 0};
   
   struct VoxIndex seed6 = {56, 62, 0};
   struct VoxIndex seed7 = {56, 59, 0};
   struct VoxIndex seed8 = {56, 55, 0};
   */
  
  /* IMPORTANT: seed positions for 128^3 version of model
   struct VoxIndex seed1 = {0, 0, 0};
   struct VoxIndex seed2 = {127, 127, 0};
   struct VoxIndex seed3 = {127, 0, 0};
   struct VoxIndex seed4 = {120, 0, 0};
   struct VoxIndex seed5 = {110, 0, 0};
   
   struct VoxIndex seed6 = {110, 125, 0};
   struct VoxIndex seed7 = {110, 120, 0};
   struct VoxIndex seed8 = {110, 110, 0};
   */
  
  /* IMPORTANT: seed positions for 256^3 version of model
   struct VoxIndex seed1 = {0, 0, 0};
   struct VoxIndex seed2 = {255, 255, 0};
   struct VoxIndex seed3 = {255, 0, 0};
   struct VoxIndex seed4 = {230, 0, 0};
   struct VoxIndex seed5 = {215, 0, 0};
   
   struct VoxIndex seed6 = {215, 250, 0};
   struct VoxIndex seed7 = {215, 242, 0};
   struct VoxIndex seed8 = {215, 215, 0};
   */
  
  /* IMPORTANT: seed positions for 512^3 version of model
   struct VoxIndex seed1 = {0, 0, 0};
   struct VoxIndex seed2 = {511, 511, 0};
   struct VoxIndex seed3 = {511, 0, 0};
   struct VoxIndex seed4 = {471, 0, 0};
   struct VoxIndex seed5 = {431, 0, 0};
   
   struct VoxIndex seed6 = {431, 500, 0};
   struct VoxIndex seed7 = {431, 482, 0};
   struct VoxIndex seed8 = {431, 452, 0};
   
  plotVoxel(0, 0, 0, 1.0f);
  plotVoxel(511, 511, 0, 1.0f);
  plotVoxel(511, 0, 0, 1.0f);
  plotVoxel(471, 0, 0, 1.0f);
  plotVoxel(431, 0, 0, 1.0f);
  plotVoxel(431, 500, 0, .0f);
  plotVoxel(431, 482, 0, 1.0f);
  plotVoxel(431, 452, 0, 1.0f);
  */
  
  
  
}

void Models::plotFIFOtest()
{
  short c = GRID_SIZE/2;
  short z = 0;
  
  for (short x = 0; x < GRID_SIZE; ++x) {
    for (short y = 0; y < GRID_SIZE; ++y) {
      
      if (pow(x-c, 2.0) + pow(y-c, 2.0) <= (double)pow(GRID_SIZE/2.7, 2.0)) {
        plotVoxel(x+(GRID_SIZE/8), y+(GRID_SIZE/8), z, 0.9f);
      }
      if (pow(x-c, 2.0) + pow(y-c, 2.0) <= (double)pow(GRID_SIZE/3.0, 2.0)) {
        plotVoxel(x+(GRID_SIZE/8), y+(GRID_SIZE/8), z, 0.81f);
      }
      if (pow(x-c, 2.0) + pow(y-c, 2.0) <= (double)pow(GRID_SIZE/4.0, 2.0)) {
        plotVoxel(x+(GRID_SIZE/8), y+(GRID_SIZE/8), z, 0.77f);
      }
    }
  }
  //plotVoxel(40, 40, 0, 1.0f);
}

void Models::plotBiasTest()
{
  //Region A
  float val = 0.28f; // 0.28
  plotVoxel(0, 0, 0, val);
  plotVoxel(0, 1, 0, val);
  plotVoxel(0, 2, 0, val);
  plotVoxel(0, 3, 0, val);
  plotVoxel(0, 4, 0, val);
  plotVoxel(0, 5, 0, val);
  plotVoxel(1, 0, 0, val);
  plotVoxel(1, 1, 0, val);
  plotVoxel(1, 2, 0, val);
  plotVoxel(1, 3, 0, val);
  plotVoxel(1, 4, 0, val);
  plotVoxel(1, 5, 0, val);
  
  //Region B
  val = 0.35f; // 0.35
  plotVoxel(2, 0, 0, val);
  plotVoxel(2, 1, 0, val);
  plotVoxel(2, 2, 0, val);
  plotVoxel(2, 3, 0, val);
  plotVoxel(2, 4, 0, val);
  plotVoxel(2, 5, 0, val);
  plotVoxel(3, 0, 0, val);
  plotVoxel(3, 1, 0, val);
  plotVoxel(3, 2, 0, val);
  plotVoxel(3, 3, 0, val);
  plotVoxel(3, 4, 0, val);
  plotVoxel(3, 5, 0, val);
  
  //Region C
  val = 0.43f; // 0.42
  plotVoxel(4, 0, 0, val);
  plotVoxel(4, 1, 0, val);
  plotVoxel(4, 2, 0, val);
  plotVoxel(4, 3, 0, val);
  plotVoxel(4, 4, 0, val);
  plotVoxel(4, 5, 0, val);
  plotVoxel(5, 0, 0, val);
  plotVoxel(5, 1, 0, val);
  plotVoxel(5, 2, 0, val);
  plotVoxel(5, 3, 0, val);
  plotVoxel(5, 4, 0, val);
  plotVoxel(5, 5, 0, val);  
}

void Models::plotSRGTest1()
{
  //Region A
  float val = 0.28f;
  plotVoxel(0, 0, 0, val);
  plotVoxel(0, 1, 0, val);
  plotVoxel(0, 2, 0, val);
  plotVoxel(1, 0, 0, val);
  plotVoxel(1, 1, 0, val);
  plotVoxel(1, 2, 0, val);
  
  //Region B
  val = 0.35f;
  plotVoxel(2, 0, 0, val);
  plotVoxel(2, 1, 0, val);
  plotVoxel(2, 2, 0, val);
  plotVoxel(3, 0, 0, val);
  plotVoxel(3, 1, 0, val);
  plotVoxel(3, 2, 0, val);
  
  //Region C
  val = 0.42f;
  plotVoxel(4, 0, 0, val);
  plotVoxel(4, 1, 0, val);
  plotVoxel(4, 2, 0, val);
  plotVoxel(5, 0, 0, val);
  plotVoxel(5, 1, 0, val);
  plotVoxel(5, 2, 0, val);
  
  floodFillNull(0.15f);
  
  
  for (int x = 0; x < GRID_SIZE; ++x) {
    for (int y = 0; y < GRID_SIZE; ++y) {
      for (int z = 1; z < GRID_SIZE; ++z) {
        deleteVoxel(x, y, z);
      }
    }
  }
  
}

void Models::plotSRGTest2()
{
  short c = GRID_SIZE/2;
  
  for (short x = 0; x < GRID_SIZE; ++x) {
    for (short y = 0; y < GRID_SIZE; ++y) {
      for (short z = 0; z < GRID_SIZE; ++z) {
        if (pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) >= (double)GRID_SIZE*GRID_SIZE/1.07 &&
            pow(x, 2.0) + pow(y, 2.0) + pow(z, 2.0) <= (double)GRID_SIZE*GRID_SIZE) {
          plotVoxel(x, y, z, 0.17f); // color, temp val
        }
        
        if (z < GRID_SIZE/4) {
          if (pow(x-c, 2.0) + pow(y-c, 2.0) <= (double)pow(GRID_SIZE/6.5, 2.0)) {
            plotVoxel(x+(GRID_SIZE/3), y+(GRID_SIZE/3), z, 0.9f);
          }
        }
      }
    }
  }
  floodFillNull(0.1);
}

void Models::floodFillNull(float intensity)
{
  for (int x = 0; x < GRID_SIZE; ++x) {
    for (int y = 0; y < GRID_SIZE; ++y) {
      for (int z = 0; z < GRID_SIZE; ++z) {
        if (voxels[x][y][z] == NULL) {
          plotVoxel(x, y, z, intensity);
        }
      }
    }
  }
}

void Models::plot3DGasket()
{
  short ind[4][3] = {{0, 0, 0}, {GRID_SIZE-1, 0, 0},
                     {GRID_SIZE/2, GRID_SIZE-1, GRID_SIZE}, // /2.7 last
                     {GRID_SIZE/2, 0, GRID_SIZE-1}};

  short point[3] = {GRID_SIZE/4, GRID_SIZE/3, GRID_SIZE/2}; // /4 /3 /2
  
  int numIter = 2000000;
  for (int i = 0; i < numIter; ++i) {
    int j = rand() % 4;
    int c = rand() % 100;
    float color = ((float)c) / 100.0;
    if (color < 0.7) color = 0.2;
    point[0] = (point[0]+ind[j][0])/2;
    point[1] = (point[1]+ind[j][1])/2;
    point[2] = (point[2]+ind[j][2])/2;
    if (!voxels[point[0]][point[1]][point[2]] && i > 0) {
      plotVoxel(point[0], point[1], point[2], color);
      //plotVoxel(point[0], point[2], point[1], 0.6f);
      //plotVoxel(point[1], point[0], point[2], 0.9f);
      //plotVoxel(point[1], point[2], point[0], 0.5f);
      //plotVoxel(point[2], point[0], point[1], 0.8f);
      plotVoxel(point[2], point[1], point[0], color);
    }
  }
}

void Models::plotGrid()
{
  for (int i = 0; i < GRID_SIZE; ++i) {
    for (int j = 0; j < GRID_SIZE; ++j) {
      for (int k = 0; k < GRID_SIZE; ++k) {
        plotVoxel(i, j, k, 0.9f);
      }
    }
  }
}

void Models::plotGridWithBorder()
{
  for (int i = 1; i < GRID_SIZE-1; ++i) {
    for (int j = 1; j < GRID_SIZE-1; ++j) {
      for (int k = 1; k < GRID_SIZE-1; ++k) {
        plotVoxel(i, j, k, 0.9f);
      }
    }
  }
}

void Models::plotBorg()
{
  for (int x = 0; x < GRID_SIZE; ++x) {
    for (int y = 0; y < GRID_SIZE; ++y) {
      for (int z = 0; z < GRID_SIZE; ++z) {
        int j = rand() % 4;
        int c = rand() % 100;
        float color = ((float)c) / 100.0;
        if (color < 0.2) color = 0.2;
        float val = sin(x*y)+sin(y*z)+sin(z*x);
        float diff = fabs(val);
        //cout << val << ", ";
        if ( diff < 0.005) {
          plotVoxel(x, y, z, color);
        }
      }
    }
  }
}


void Models::initVoxelGrid()
{
  voxels.resize(GRID_SIZE);
  for (int i = 0; i < GRID_SIZE; ++i) {
    voxels[i].resize(GRID_SIZE);
    for (int j = 0; j < GRID_SIZE; ++j) {
      voxels[i][j].resize(GRID_SIZE);
    }
  }
  
  for (short i = 0; i < GRID_SIZE; ++i) {
    for (short j = 0; j < GRID_SIZE; ++j) {
      for (short k = 0; k < GRID_SIZE; ++k) {
        voxels[i][j][k] = NULL;
      }
    }
  }
}

void Models::initSecVoxelGrid()
{
  sec_voxels.resize(GRID_SIZE);
  for (int i = 0; i < GRID_SIZE; ++i) {
    sec_voxels[i].resize(GRID_SIZE);
    for (int j = 0; j < GRID_SIZE; ++j) {
      sec_voxels[i][j].resize(GRID_SIZE);
    }
  }
  
  for (short i = 0; i < GRID_SIZE; ++i) {
    for (short j = 0; j < GRID_SIZE; ++j) {
      for (short k = 0; k < GRID_SIZE; ++k) {
        sec_voxels[i][j][i] = NULL;
      }
    }
  }
  secVoxelsExist = true;
}

void Models::booleanSubtract()
{
  if (!secVoxelsExist) return;
  for (short i = 0; i < GRID_SIZE; ++i) {
    for (short j = 0; j < GRID_SIZE; ++j) {
      for (short k = 0; k < GRID_SIZE; ++k) {
        if (voxels[i][j][k] && sec_voxels[i][j][k]) {
          delete voxels[i][j][k];
          voxels[i][j][k] = NULL;
        }
      }
    }
  }
}

void Models::plotSecSphere()
{
  short c = GRID_SIZE/2;
  for (short x = 0; x < GRID_SIZE; ++x) {
    for (short y = 0; y < GRID_SIZE; ++y) {
      for (short z = 0; z < GRID_SIZE; ++z) {
        if (pow(x-c, 2.0) + pow(y-c, 2.0) + pow(z-c, 2.0) <= (double)pow((GRID_SIZE/2.2), 2.0)) {
          int randNum = rand() % 40 +  35;
          float color = randNum * 0.01;
          //cout << x+GRID_SIZE << ", " << y+GRID_SIZE << ", " << z+GRID_SIZE << endl;
          sec_voxels[x][y][z] = new Voxel(0.2f); // color, temp val
        }
      }
    }
  }
}

vector<short> Models::createVoxelIndex(short x, short y, short z)
{
  vector<short> indexVector;
  indexVector.push_back(x);
  indexVector.push_back(y);
  indexVector.push_back(z);
  return indexVector;
}

void Models::setRenderVoxels(vector<vector<short> > r_vox_ind)
{
  drawVoxelsEnabled = false;
  vector<Voxel*> renderVoxelsTemp;
  for (int i = 0; i < r_vox_ind.size(); ++i) {
    renderVoxelsTemp.push_back(voxels[r_vox_ind[i][0]][r_vox_ind[i][1]][r_vox_ind[i][2]]);
  }
  renderVoxels = renderVoxelsTemp;
  renderVoxelsIndex = r_vox_ind;
  drawVoxelsEnabled = true;
}

void Models::updateRenderVoxels()
{
  drawVoxelsEnabled = false;
  vector<Voxel*> renderVoxelsTemp;
  vector<vector<short> > renderVoxelsIndexTemp;
  int numVoxels = 0;
  for (short i = 0; i < GRID_SIZE; ++i) {
    for (short j = 0; j < GRID_SIZE; ++j) {
      for (short k = 0; k < GRID_SIZE; ++k) {
        numVoxels++;
        if (voxels[i][j][k]) {
          if (i == 0 || j == 0 || k == 0 || i == GRID_SIZE-1 || j == GRID_SIZE-1 || k == GRID_SIZE-1) { // always render border voxels
            renderVoxelsTemp.push_back(voxels[i][j][k]); // store pointer to voxel
            renderVoxelsIndexTemp.push_back(createVoxelIndex(i,j,k)); // store voxel index
          } 
          else {
            if (voxels[i+1][j][k] && voxels[i-1][j][k] &&
                voxels[i][j+1][k] && voxels[i][j-1][k] &&
                voxels[i][j][k+1] && voxels[i][j][k-1]) {
              // not visible, do not render
            } else { // render visible voxel
              renderVoxelsTemp.push_back(voxels[i][j][k]); // store pointer to voxel
              renderVoxelsIndexTemp.push_back(createVoxelIndex(i,j,k)); // store voxel index
            }
          }
        }
      }
    }
  }
  renderVoxels = renderVoxelsTemp;
  renderVoxelsIndex = renderVoxelsIndexTemp;
  drawVoxelsEnabled = true;
}

void Models::plotVoxel(int x, int y, int z, float val)
{
  if (x > GRID_SIZE-1 || y > GRID_SIZE-1 || z > GRID_SIZE-1) {
    cout << "ERROR: Invalid grid coordinates.";
    return;
  }
  if (voxels[x][y][z]) voxels[x][y][z]->setIntensity(val);
  else voxels[x][y][z] = new Voxel(val);
}

void Models::deleteVoxel(int x, int y, int z)
{
  if (voxels[x][y][z]) {
    delete voxels[x][y][z];
    voxels[x][y][z] = NULL;
  }
}

void Models::deleteAllVoxels()
{
  for (short i = 0; i < GRID_SIZE; ++i) {
    for (short j = 0; j < GRID_SIZE; ++j) {
      for (short k = 0; k < GRID_SIZE; ++k) {
        if (voxels[i][j][k]) {
          delete voxels[i][j][k];
          voxels[i][j][k] = NULL;
        }
        if (sec_voxels[i][j][k]) {
          delete sec_voxels[i][j][k];
          sec_voxels[i][j][k] = NULL;
        }
      }
    }
  }
  updateRenderVoxels();
}
