// Segmentation.cpp

#include "Segmentation.h"
#include "Voxel.h"
#include "VoxIndex.h"
#include "SSLElem.h"
#include "Globals.h"
#include <vector>
#include <deque>
#include <list>
#include <queue>
#include <set>
#include <cmath>
#include <iostream>

using std::cout;
using std::endl;

short Segmentation::neighborhood = 6;
const short Segmentation::MIN_REGION_SIZE = 700; // 60 // 120 // 700 (medica)
unsigned short Segmentation::currentLabel = 1;
unsigned short Segmentation::currentlySelectedSegment = -1;
float Segmentation::sslDeltaCache = 0.0f;
float Segmentation::sslNumDeltasCache = 0.0f;
multiset<SSLElem>::iterator Segmentation::sslSearchHeuristic = 0;
bool Segmentation::sslSearchHeuristicInitialized = false;
bool Segmentation::sslHeuristicSearchOn = true;
bool Segmentation::renderSegmentSurface = true;
vector<vector<vector<short> > > Segmentation::segments;
vector<vector<vector<short> > > Segmentation::segmentSurfaces;
vector<vector<short> > Segmentation::currentRegion;
vector<vector<short> > Segmentation::currentRegionSurface;


Segmentation::Segmentation() {}

Segmentation::~Segmentation() {}

void Segmentation::addVoxelToCurrentRegion(short x, short y, short z)
{
  vector<short> voxIndex;
  voxIndex.push_back(x);
  voxIndex.push_back(y);
  voxIndex.push_back(z);
  currentRegion.push_back(voxIndex);
}

void Segmentation::addVoxelToCurrentRegionSurface(short x, short y, short z)
{
  vector<short> voxIndex;
  voxIndex.push_back(x);
  voxIndex.push_back(y);
  voxIndex.push_back(z);
  currentRegionSurface.push_back(voxIndex);
}

bool Segmentation::intensitiesAreEqual(float in1, float in2)
{
  static const float margin = 0.3; // 0.3 (medical) 0.18 (medical) // 0.15 (medical) // 0.1 (medical) // 0.08 // 0.02 // 0.005
  float diff = fabs(in1 - in2);
  if (diff > margin) return false; // intensity of voxel is not equal to target intensity
  else return true;
}

bool Segmentation::floodFillRegionHelper(short x, short y, short z, vector<vector<short> >& queue,
                                         float& tIntensity)
{
  if (x < GRID_SIZE && y < GRID_SIZE && z < GRID_SIZE && x >= 0 && y >= 0 && z >= 0) {
    if (voxels[x][y][z] == NULL) return true;
    if (voxels[x][y][z]->getLabel() != UNLABELED) {
      if (voxels[x][y][z]->getLabel() == currentLabel) return false;
      else return true;
    }
    if (!intensitiesAreEqual(voxels[x][y][z]->getIntensity(), tIntensity)) {
      return true;
    }
    //if (voxels[x][y][z]->getLabel() != UNLABELED) return;
    //if (intensitiesAreEqual(voxels[x][y][z]->getIntensity(), tIntensity)) {
    else {
      //this is the next voxel in the current region
      addVoxelToCurrentRegion(x, y, z);
      voxels[x][y][z]->setLabel(currentLabel); // set temp label
      vector<short> vi;
      vi.push_back(x);
      vi.push_back(y);
      vi.push_back(z);
      queue.push_back(vi);
      return false;
    }
  }
  return true;
}

void Segmentation::floodFillRegion(short x, short y, short z, float tIntensity)
{
  //ftime(&start_time); // function timing
  if (voxels[x][y][z] == NULL) return;
  
  // this is the first voxel in the new region
  addVoxelToCurrentRegion(x, y, z);
  addVoxelToCurrentRegionSurface(x, y, z);
  voxels[x][y][z]->setLabel(currentLabel); // set temp label
  
  vector<vector<short> > queue;
  
  vector<short> vi;
  vi.push_back(x);
  vi.push_back(y);
  vi.push_back(z);
  queue.push_back(vi);
  
  while (!queue.empty()) {
    //display();
    
    vector<short> n = queue.back();
    
    queue.pop_back();
    
    // 6-connected (surface connected)
    bool s1 = floodFillRegionHelper(n.at(0)+1, n.at(1), n.at(2), queue, tIntensity);
    bool s2 = floodFillRegionHelper(n.at(0)-1, n.at(1), n.at(2), queue, tIntensity);
    bool s3 = floodFillRegionHelper(n.at(0), n.at(1)+1, n.at(2), queue, tIntensity);
    bool s4 = floodFillRegionHelper(n.at(0), n.at(1)-1, n.at(2), queue, tIntensity);
    bool s5 = floodFillRegionHelper(n.at(0), n.at(1), n.at(2)+1, queue, tIntensity);
    bool s6 = floodFillRegionHelper(n.at(0), n.at(1), n.at(2)-1, queue, tIntensity);
    //cout << "surface values:\n";
    //cout << "s1: " << s1 << ", s2: " << s2 << ", s3: " << s3 << ", s4: " << s4 << ", s5: " << s5 << ", s6: " << s6 << endl;
    if (s1 || s2 || s3 || s4 || s5 || s6) {
      //cout << "Adding voxel " << n[0] << ", " << n[1] << ", " << n[2] << " to surface region\n";
      addVoxelToCurrentRegionSurface(n[0], n[1], n[2]);
    }
    // 14-connected (edge connected)
    floodFillRegionHelper(n.at(0)+1, n.at(1)+1, n.at(2), queue, tIntensity);
    floodFillRegionHelper(n.at(0)-1, n.at(1)-1, n.at(2), queue, tIntensity);
    floodFillRegionHelper(n.at(0)+1, n.at(1)-1, n.at(2), queue, tIntensity);
    floodFillRegionHelper(n.at(0)-1, n.at(1)+1, n.at(2), queue, tIntensity);
    floodFillRegionHelper(n.at(0), n.at(1)+1, n.at(2)+1, queue, tIntensity);
    floodFillRegionHelper(n.at(0), n.at(1)-1, n.at(2)-1, queue, tIntensity);
    floodFillRegionHelper(n.at(0), n.at(1)+1, n.at(2)-1, queue, tIntensity);
    floodFillRegionHelper(n.at(0), n.at(1)-1, n.at(2)+1, queue, tIntensity);

  }
  
  if (currentRegion.size() < MIN_REGION_SIZE) { // noise has been found
    for (int i = 0; i < currentRegion.size(); ++i) {
      short x_ind = currentRegion.at(i).at(0);
      short y_ind = currentRegion.at(i).at(1);
      short z_ind = currentRegion.at(i).at(2);
      voxels[x_ind][y_ind][z_ind]->setLabel(NOISE);
    }
  } else { // a new region has been found
    segments.push_back(currentRegion);
    segmentSurfaces.push_back(currentRegionSurface);
    currentLabel++;
  }
  
  //cleanup
  currentRegion.clear();
  currentRegionSurface.clear();
  
  
  // calculate running time of function
  //ftime(&current_time);
  //ms_since_start = (long) (1000.0 * (current_time.time - start_time.time) + (current_time.millitm - start_time.millitm));
  //printf("ms since start: %d \n", ms_since_start);
}

void Segmentation::resetDataStructures()
{
  currentLabel = 1;
  currentlySelectedSegment = -1;
  segments.clear();
  segmentSurfaces.clear();
  currentRegion.clear();
  currentRegionSurface.clear();
}

void Segmentation::growRegions()
{
  resetDataStructures();
  for (int x = 0; x < GRID_SIZE; ++x) {
    for (int y = 0; y < GRID_SIZE; ++y) {
      for (int z = 0; z < GRID_SIZE; ++z) {
        if (currentLabel == NOISE || currentLabel == UNLABELED) {
          cout << "No more available labels!\n";
          return;
        }
        if (voxels[x][y][z] == NULL) continue; // if voxel is empty (background), skip it
        //cout << "voxel label: " << voxels[x][y][z]->getLabel() << endl;
        if (voxels[x][y][z]->getLabel() == UNLABELED) { // if voxel is unlabeled, start region growing
          floodFillRegion(x, y, z, voxels[x][y][z]->getIntensity());
        }
      }
    }
  }
  cout << "Number of labels: " << currentLabel << endl;
}

//--------------------------------------------------------------------------------------

void Segmentation::growRegionsS3DRG(void (*display)(void), short neighbors)
{
  resetDataStructures();
  neighborhood = neighbors;
  for (int x = 0; x < GRID_SIZE; ++x) {
    for (int y = 0; y < GRID_SIZE; ++y) {
      for (int z = 0; z < GRID_SIZE; ++z) {
        if (currentLabel == NOISE || currentLabel == UNLABELED) {
          cout << "No more available labels!\n";
          return;
        }
        if (voxels[x][y][z] == NULL) continue; // if voxel is empty (background), skip it
        if (voxels[x][y][z]->getLabel() == UNLABELED) { // if voxel is unlabeled, start region growing
          floodFillRegionS3DRG(x, y, z, voxels[x][y][z]->getIntensity(), display);
        }
      }
    }
  }
  cout << "Number of labels: " << currentLabel << endl;
}

void Segmentation::floodFillRegionS3DRG(short x, short y, short z, float tIntensity, void (*display)(void))
{
  if (voxels[x][y][z] == NULL) return;
  
  // this is the first voxel in the new region
  addVoxelToCurrentRegion(x, y, z);
  addVoxelToCurrentRegionSurface(x, y, z);
  voxels[x][y][z]->setLabel(currentLabel); // set temp label
  
  deque<vector<short> > queue;
  
  vector<short> vi;
  vi.push_back(x);
  vi.push_back(y);
  vi.push_back(z);
  queue.push_back(vi);
  
  int counter = 1;
  
  while (!queue.empty()) {

    //int randNum = rand() % 100;
    //if (randNum == 0) display();
    //display();
    
    // LIFO queue
    //vector<short> n = queue.back();
    //queue.pop_back();
    
    // FIFO queue
    vector<short> n = queue.front();
    queue.erase(queue.begin());
    
    // 6-connected (surface)
    bool s1 = floodFillRegionHelperS3DRG(n.at(0)+1, n.at(1), n.at(2), queue, tIntensity);
    bool s2 = floodFillRegionHelperS3DRG(n.at(0)-1, n.at(1), n.at(2), queue, tIntensity);
    bool s3 = floodFillRegionHelperS3DRG(n.at(0), n.at(1)+1, n.at(2), queue, tIntensity);
    bool s4 = floodFillRegionHelperS3DRG(n.at(0), n.at(1)-1, n.at(2), queue, tIntensity);
    bool s5 = floodFillRegionHelperS3DRG(n.at(0), n.at(1), n.at(2)+1, queue, tIntensity);
    bool s6 = floodFillRegionHelperS3DRG(n.at(0), n.at(1), n.at(2)-1, queue, tIntensity);
    if (s1 || s2 || s3 || s4 || s5 || s6) addVoxelToCurrentRegionSurface(n[0], n[1], n[2]);
    
    if (neighborhood == 26) {
      // 18-connected (surface + edge)
      floodFillRegionHelperS3DRG(n.at(0)+1, n.at(1)+1, n.at(2), queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)-1, n.at(1)-1, n.at(2), queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)+1, n.at(1)-1, n.at(2), queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)-1, n.at(1)+1, n.at(2), queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0), n.at(1)+1, n.at(2)+1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0), n.at(1)-1, n.at(2)-1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0), n.at(1)+1, n.at(2)-1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0), n.at(1)-1, n.at(2)+1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)+1, n.at(1), n.at(2)+1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)-1, n.at(1), n.at(2)-1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)+1, n.at(1), n.at(2)-1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)-1, n.at(1), n.at(2)+1, queue, tIntensity);
      
      // 26-connected (surface + edge + vertex)
      floodFillRegionHelperS3DRG(n.at(0)+1, n.at(1)+1, n.at(2)+1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)+1, n.at(1)+1, n.at(2)-1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)-1, n.at(1)+1, n.at(2)+1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)-1, n.at(1)+1, n.at(2)-1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)+1, n.at(1)-1, n.at(2)+1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)+1, n.at(1)-1, n.at(2)-1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)-1, n.at(1)-1, n.at(2)+1, queue, tIntensity);
      floodFillRegionHelperS3DRG(n.at(0)-1, n.at(1)-1, n.at(2)-1, queue, tIntensity);
    }
    //if (counter == 120) break;
    counter++;
  }

  if (currentRegion.size() < MIN_REGION_SIZE) { // noise has been found
    for (int i = 0; i < currentRegion.size(); ++i) {
      short x_ind = currentRegion.at(i).at(0);
      short y_ind = currentRegion.at(i).at(1);
      short z_ind = currentRegion.at(i).at(2);
      voxels[x_ind][y_ind][z_ind]->setLabel(NOISE);
    }
  } else { // a new region has been found
    segments.push_back(currentRegion);
    segmentSurfaces.push_back(currentRegionSurface);
    currentLabel++;
  }
   
  //currentLabel++;
   
  //cleanup
  currentRegion.clear();
  currentRegionSurface.clear();
  
}

bool Segmentation::floodFillRegionHelperS3DRG(short x, short y, short z, deque<vector<short> >& queue,
                                         float& tIntensity)
{
  if (x < GRID_SIZE && y < GRID_SIZE && z < GRID_SIZE && x >= 0 && y >= 0 && z >= 0) {
    if (voxels[x][y][z] == NULL) return true;
    if (voxels[x][y][z]->getLabel() != UNLABELED) {
      if (voxels[x][y][z]->getLabel() == currentLabel) return false;
      else return true;
    }
    if (!intensitiesAreEqual(voxels[x][y][z]->getIntensity(), tIntensity)) {
      return true;
    }
    else {
      //this is the next voxel in the current region
      addVoxelToCurrentRegion(x, y, z);
      voxels[x][y][z]->setLabel(currentLabel); // set temp label
      vector<short> vi;
      vi.push_back(x);
      vi.push_back(y);
      vi.push_back(z);
      queue.push_back(vi);
      return false;
    }
  }
  return true;
}

//--------------------------------------------------------------------------------------

void Segmentation::growRegionsS2DRG(short neighbors)
{
  resetDataStructures();
  neighborhood = neighbors;
  for (int x = 0; x < GRID_SIZE; ++x) {
    for (int y = 0; y < GRID_SIZE; ++y) {
      if (currentLabel == NOISE || currentLabel == UNLABELED) {
        cout << "No more available labels!\n";
        return;
      }
      if (voxels[x][y][0] == NULL) continue; // if voxel is empty (background), skip it
      if (voxels[x][y][0]->getLabel() == UNLABELED) { // if voxel is unlabeled, start region growing
        floodFillRegionS2DRG(x, y, 0, voxels[x][y][0]->getIntensity());
      }
    }
  }
  cout << "Number of labels: " << currentLabel << endl;
}

void Segmentation::floodFillRegionS2DRG(short x, short y, short z, float tIntensity)
{
  if (voxels[x][y][z] == NULL) return;
  
  // this is the first voxel in the new region
  addVoxelToCurrentRegion(x, y, z);
  addVoxelToCurrentRegionSurface(x, y, z);
  voxels[x][y][z]->setLabel(currentLabel); // set temp label
  
  vector<vector<short> > queue;
  
  vector<short> vi;
  vi.push_back(x);
  vi.push_back(y);
  vi.push_back(z);
  queue.push_back(vi);
  
  while (!queue.empty()) {
    //display();
    
    vector<short> n = queue.back();
    
    queue.pop_back();
    
    // 4-connected (edge)
    bool s1 = floodFillRegionHelperS2DRG(n.at(0)+1, n.at(1), n.at(2), queue, tIntensity);
    bool s2 = floodFillRegionHelperS2DRG(n.at(0)-1, n.at(1), n.at(2), queue, tIntensity);
    bool s3 = floodFillRegionHelperS2DRG(n.at(0), n.at(1)+1, n.at(2), queue, tIntensity);
    bool s4 = floodFillRegionHelperS2DRG(n.at(0), n.at(1)-1, n.at(2), queue, tIntensity);
    if (s1 || s2 || s3 || s4) addVoxelToCurrentRegionSurface(n[0], n[1], n[2]);
    
    if (neighborhood == 8) {
      // 8-connected (edge + vertex)
      floodFillRegionHelperS2DRG(n.at(0)+1, n.at(1)+1, n.at(2), queue, tIntensity);
      floodFillRegionHelperS2DRG(n.at(0)-1, n.at(1)-1, n.at(2), queue, tIntensity);
      floodFillRegionHelperS2DRG(n.at(0)+1, n.at(1)-1, n.at(2), queue, tIntensity);
      floodFillRegionHelperS2DRG(n.at(0)-1, n.at(1)+1, n.at(2), queue, tIntensity);
    }
  }
  
  if (currentRegion.size() < MIN_REGION_SIZE) { // noise has been found
    for (int i = 0; i < currentRegion.size(); ++i) {
      short x_ind = currentRegion.at(i).at(0);
      short y_ind = currentRegion.at(i).at(1);
      short z_ind = currentRegion.at(i).at(2);
      voxels[x_ind][y_ind][z_ind]->setLabel(NOISE);
    }
  } else { // a new region has been found
    segments.push_back(currentRegion);
    segmentSurfaces.push_back(currentRegionSurface);
    currentLabel++;
  }
  
  //cleanup
  currentRegion.clear();
  currentRegionSurface.clear();
  
}

bool Segmentation::floodFillRegionHelperS2DRG(short x, short y, short z, vector<vector<short> >& queue,
                                              float& tIntensity)
{
  if (x < GRID_SIZE && y < GRID_SIZE && z < GRID_SIZE && x >= 0 && y >= 0 && z >= 0) {
    if (voxels[x][y][z] == NULL) return true;
    if (voxels[x][y][z]->getLabel() != UNLABELED) {
      if (voxels[x][y][z]->getLabel() == currentLabel) return false;
      else return true;
    }
    if (!intensitiesAreEqual(voxels[x][y][z]->getIntensity(), tIntensity)) {
      return true;
    }
    else {
      //this is the next voxel in the current region
      addVoxelToCurrentRegion(x, y, z);
      voxels[x][y][z]->setLabel(currentLabel); // set temp label
      vector<short> vi;
      vi.push_back(x);
      vi.push_back(y);
      vi.push_back(z);
      queue.push_back(vi);
      return false;
    }
  }
  return true;
}

//------------------------------------------------------------------------------

void Segmentation::printSetInfo(set<int> s)
{
  cout << endl;
  cout << "Set information: " << endl;
  cout << "-----------------" << endl;
  cout << "Set size: " << s.size() << endl;
  cout << "Values: ";
  set<int>::iterator it;
  for (it = s.begin(); it != s.end(); ++it) {
    cout << " " << *it;
  }
}

void Segmentation::printSSL(multiset<SSLElem>& ssl)
{
  multiset<SSLElem>::iterator sslIt;
  int counter = 1;
  for (sslIt = ssl.begin(); sslIt != ssl.end(); ++sslIt) {
    cout << "Element " << counter << endl;
    cout << "x: " << sslIt->x() << ", y: " << sslIt->y() << ", z: " << sslIt->z() << endl;
    cout << "delta: " << sslIt->delta() << endl;
    cout << "--------------------------\n";
    counter++;
  }
}

void Segmentation::insertSSLElem_List(SSLElem elem, list<SSLElem>& ssl)
{
  int label = voxels[elem.x()][elem.y()][elem.z()]->getLabel();
  if (label != UNLABELED) return;
  if (ssl.empty()) ssl.push_back(elem);
  else {
    //short curX = elem.x();
    //short curY = elem.y();
    //short curZ = elem.z();
    float curDelta = elem.delta();
    sslDeltaCache += curDelta;
    sslNumDeltasCache += 1;
    float average = sslDeltaCache/sslNumDeltasCache;
    //cout << "Average: " << average << endl;
    list<SSLElem>::iterator insertPos = ssl.end();
    list<SSLElem>::iterator sslIt;
    //bool posUpdated = false;
    if (curDelta <= average) {
      for (sslIt = ssl.begin(); sslIt != ssl.end(); ++sslIt) {
        //if (curX == sslIt->x() && curY == sslIt->y() && curZ == sslIt->z()) return; // found duplicate
        if (curDelta <= sslIt->delta() /*&& posUpdated == false*/) {
          insertPos = sslIt;
          //posUpdated = true;
          break;
          
        }
      }
    } else {
      for (sslIt = ssl.end(); sslIt != ssl.begin(); --sslIt) {
        if (curDelta >= sslIt->delta()) {
          insertPos = sslIt;
          break;
        }
      }
    }
    voxels[elem.x()][elem.y()][elem.z()]->setLabel(SSL_ELEM);
    ssl.insert(insertPos, elem);
    
  }
}

void Segmentation::insertSSLElem_Heap(SSLElem elem, reservable_priority_queue<SSLElem>& ssl)
{
  int label = voxels[elem.x()][elem.y()][elem.z()]->getLabel();
  if (label != UNLABELED) return;
  ssl.push(elem);
  voxels[elem.x()][elem.y()][elem.z()]->setLabel(SSL_ELEM);
}

void Segmentation::insertSSLElem_BinarySearchTree(SSLElem elem, multiset<SSLElem>& ssl)
{
  int label = voxels[elem.x()][elem.y()][elem.z()]->getLabel();
  if (label != UNLABELED) return;
  if (sslSearchHeuristicInitialized && sslHeuristicSearchOn) {
    sslSearchHeuristic = ssl.insert(sslSearchHeuristic, elem);
  } else {
    sslSearchHeuristic = ssl.insert(elem);
  }
  voxels[elem.x()][elem.y()][elem.z()]->setLabel(SSL_ELEM);
}

void Segmentation::addNeighborToSSL_Heap(short x, short y, short z, reservable_priority_queue<SSLElem>& ssl, float regionMean)
{
  if (validIndex(x, y, z)) {
    float intensity = voxels[x][y][z]->getIntensity();
    SSLElem elem(x, y, z, getDelta(intensity, regionMean));
    insertSSLElem_Heap(elem, ssl);
  }
}

void Segmentation::addNeighborToSSL_BinarySearchTree(short x, short y, short z, multiset<SSLElem>& ssl, float regionMean)
{
  if (validIndex(x, y, z)) {
    float intensity = voxels[x][y][z]->getIntensity();
    SSLElem elem(x, y, z, getDelta(intensity, regionMean));
    insertSSLElem_BinarySearchTree(elem, ssl);
  }
}

void Segmentation::addNeighborToSSL_List(short x, short y, short z, list<SSLElem>& ssl, float regionMean)
{
  if (validIndex(x, y, z)) {
    float intensity = voxels[x][y][z]->getIntensity();
    SSLElem elem(x, y, z, getDelta(intensity, regionMean));
    insertSSLElem_List(elem, ssl); // doubly-linked list
  }
}

void Segmentation::add6NeighborsToSSL_DualHeap(struct VoxIndex vi,
                                               reservable_priority_queue<SSLElem>& ssl,
                                               reservable_priority_queue<SSLElem>& ssl2,
                                               float regionMean)
{
  // heap 1
  addNeighborToSSL_Heap(vi.x-1, vi.y, vi.z, ssl, regionMean); // left
  addNeighborToSSL_Heap(vi.x+1, vi.y, vi.z, ssl, regionMean); // right
  addNeighborToSSL_Heap(vi.x, vi.y-1, vi.z, ssl, regionMean); // down
  
  // heap 2
  addNeighborToSSL_Heap(vi.x, vi.y+1, vi.z, ssl2, regionMean); // up
  addNeighborToSSL_Heap(vi.x, vi.y, vi.z-1, ssl2, regionMean); // out
  addNeighborToSSL_Heap(vi.x, vi.y, vi.z+1, ssl2, regionMean); // in
}

void Segmentation::add6NeighborsToSSL_Heap(struct VoxIndex vi, reservable_priority_queue<SSLElem>& ssl, float regionMean)
{
  addNeighborToSSL_Heap(vi.x-1, vi.y, vi.z, ssl, regionMean); // left
  addNeighborToSSL_Heap(vi.x+1, vi.y, vi.z, ssl, regionMean); // right
  addNeighborToSSL_Heap(vi.x, vi.y-1, vi.z, ssl, regionMean); // down
  addNeighborToSSL_Heap(vi.x, vi.y+1, vi.z, ssl, regionMean); // up
  addNeighborToSSL_Heap(vi.x, vi.y, vi.z-1, ssl, regionMean); // out
  addNeighborToSSL_Heap(vi.x, vi.y, vi.z+1, ssl, regionMean); // in
}

void Segmentation::add6NeighborsToSSL_BinarySearchTree(struct VoxIndex vi, multiset<SSLElem>& ssl, float regionMean)
{
  addNeighborToSSL_BinarySearchTree(vi.x-1, vi.y, vi.z, ssl, regionMean); // left
  addNeighborToSSL_BinarySearchTree(vi.x+1, vi.y, vi.z, ssl, regionMean); // right
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y-1, vi.z, ssl, regionMean); // down
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y+1, vi.z, ssl, regionMean); // up
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y, vi.z-1, ssl, regionMean); // out
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y, vi.z+1, ssl, regionMean); // in
}

void Segmentation::add6NeighborsToSSL_List(struct VoxIndex vi, list<SSLElem>& ssl, float regionMean)
{
  addNeighborToSSL_List(vi.x-1, vi.y, vi.z, ssl, regionMean); // left
  addNeighborToSSL_List(vi.x+1, vi.y, vi.z, ssl, regionMean); // right
  addNeighborToSSL_List(vi.x, vi.y-1, vi.z, ssl, regionMean); // down
  addNeighborToSSL_List(vi.x, vi.y+1, vi.z, ssl, regionMean); // up
  addNeighborToSSL_List(vi.x, vi.y, vi.z-1, ssl, regionMean); // out
  addNeighborToSSL_List(vi.x, vi.y, vi.z+1, ssl, regionMean); // in
}

void Segmentation::add26NeighborsToSSL_Heap(struct VoxIndex vi, reservable_priority_queue<SSLElem>& ssl, float regionMean)
{
  addNeighborToSSL_Heap(vi.x-1, vi.y, vi.z, ssl, regionMean); // left
  addNeighborToSSL_Heap(vi.x+1, vi.y, vi.z, ssl, regionMean); // right
  addNeighborToSSL_Heap(vi.x, vi.y-1, vi.z, ssl, regionMean); // down
  addNeighborToSSL_Heap(vi.x, vi.y+1, vi.z, ssl, regionMean); // up
  addNeighborToSSL_Heap(vi.x, vi.y, vi.z-1, ssl, regionMean); // out
  addNeighborToSSL_Heap(vi.x, vi.y, vi.z+1, ssl, regionMean); // in
  
  addNeighborToSSL_Heap(vi.x+1, vi.y+1, vi.z, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x-1, vi.y-1, vi.z, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x+1, vi.y-1, vi.z, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x-1, vi.y+1, vi.z, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x, vi.y+1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x, vi.y-1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x, vi.y+1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x, vi.y-1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x+1, vi.y, vi.z+1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x-1, vi.y, vi.z-1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x+1, vi.y, vi.z-1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x-1, vi.y, vi.z+1, ssl, regionMean);
  
  addNeighborToSSL_Heap(vi.x+1, vi.y+1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x+1, vi.y+1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x-1, vi.y+1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x-1, vi.y+1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x+1, vi.y-1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x+1, vi.y-1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x-1, vi.y-1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_Heap(vi.x-1, vi.y-1, vi.z-1, ssl, regionMean);
}

void Segmentation::add26NeighborsToSSL_BinarySearchTree(struct VoxIndex vi, multiset<SSLElem>& ssl, float regionMean)
{
  addNeighborToSSL_BinarySearchTree(vi.x-1, vi.y, vi.z, ssl, regionMean); // left
  addNeighborToSSL_BinarySearchTree(vi.x+1, vi.y, vi.z, ssl, regionMean); // right
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y-1, vi.z, ssl, regionMean); // down
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y+1, vi.z, ssl, regionMean); // up
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y, vi.z-1, ssl, regionMean); // out
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y, vi.z+1, ssl, regionMean); // in
  
  addNeighborToSSL_BinarySearchTree(vi.x+1, vi.y+1, vi.z, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x-1, vi.y-1, vi.z, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x+1, vi.y-1, vi.z, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x-1, vi.y+1, vi.z, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y+1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y-1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y+1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x, vi.y-1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x+1, vi.y, vi.z+1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x-1, vi.y, vi.z-1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x+1, vi.y, vi.z-1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x-1, vi.y, vi.z+1, ssl, regionMean);
  
  addNeighborToSSL_BinarySearchTree(vi.x+1, vi.y+1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x+1, vi.y+1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x-1, vi.y+1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x-1, vi.y+1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x+1, vi.y-1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x+1, vi.y-1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x-1, vi.y-1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_BinarySearchTree(vi.x-1, vi.y-1, vi.z-1, ssl, regionMean);
}

void Segmentation::add26NeighborsToSSL_List(struct VoxIndex vi, list<SSLElem>& ssl, float regionMean)
{
  addNeighborToSSL_List(vi.x-1, vi.y, vi.z, ssl, regionMean); // left
  addNeighborToSSL_List(vi.x+1, vi.y, vi.z, ssl, regionMean); // right
  addNeighborToSSL_List(vi.x, vi.y-1, vi.z, ssl, regionMean); // down
  addNeighborToSSL_List(vi.x, vi.y+1, vi.z, ssl, regionMean); // up
  addNeighborToSSL_List(vi.x, vi.y, vi.z-1, ssl, regionMean); // out
  addNeighborToSSL_List(vi.x, vi.y, vi.z+1, ssl, regionMean); // in
  
  addNeighborToSSL_List(vi.x+1, vi.y+1, vi.z, ssl, regionMean);
  addNeighborToSSL_List(vi.x-1, vi.y-1, vi.z, ssl, regionMean);
  addNeighborToSSL_List(vi.x+1, vi.y-1, vi.z, ssl, regionMean);
  addNeighborToSSL_List(vi.x-1, vi.y+1, vi.z, ssl, regionMean);
  addNeighborToSSL_List(vi.x, vi.y+1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_List(vi.x, vi.y-1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_List(vi.x, vi.y+1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_List(vi.x, vi.y-1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_List(vi.x+1, vi.y, vi.z+1, ssl, regionMean);
  addNeighborToSSL_List(vi.x-1, vi.y, vi.z-1, ssl, regionMean);
  addNeighborToSSL_List(vi.x+1, vi.y, vi.z-1, ssl, regionMean);
  addNeighborToSSL_List(vi.x-1, vi.y, vi.z+1, ssl, regionMean);
  
  addNeighborToSSL_List(vi.x+1, vi.y+1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_List(vi.x+1, vi.y+1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_List(vi.x-1, vi.y+1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_List(vi.x-1, vi.y+1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_List(vi.x+1, vi.y-1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_List(vi.x+1, vi.y-1, vi.z-1, ssl, regionMean);
  addNeighborToSSL_List(vi.x-1, vi.y-1, vi.z+1, ssl, regionMean);
  addNeighborToSSL_List(vi.x-1, vi.y-1, vi.z-1, ssl, regionMean);
}

int Segmentation::getLabel(short x, short y, short z)
{
  if (validIndex(x, y, z)) {
    return voxels[x][y][z]->getLabel();
  } else {
    return UNLABELED;
  }
}

set<int> Segmentation::get6NeighborLabels(SSLElem& elem)
{
  set<int> result;
  int label1 = getLabel(elem.x()-1, elem.y(), elem.z());
  if (label1 != UNLABELED && label1 != SSL_ELEM) result.insert(label1);
  int label2 = getLabel(elem.x()+1, elem.y(), elem.z());
  if (label2 != UNLABELED && label2 != SSL_ELEM) result.insert(label2);
  int label3 = getLabel(elem.x(), elem.y()-1, elem.z());
  if (label3 != UNLABELED && label3 != SSL_ELEM) result.insert(label3);
  int label4 = getLabel(elem.x(), elem.y()+1, elem.z());
  if (label4 != UNLABELED && label4 != SSL_ELEM) result.insert(label4);
  int label5 = getLabel(elem.x(), elem.y(), elem.z()-1);
  if (label5 != UNLABELED && label5 != SSL_ELEM) result.insert(label5);
  int label6 = getLabel(elem.x(), elem.y(), elem.z()+1);
  if (label6 != UNLABELED && label6 != SSL_ELEM) result.insert(label6);
  return result;
}

set<int> Segmentation::get26NeighborLabels(SSLElem& elem)
{
  set<int> result;
  int label1 = getLabel(elem.x()-1, elem.y(), elem.z());
  if (label1 != UNLABELED && label1 != SSL_ELEM) result.insert(label1);
  int label2 = getLabel(elem.x()+1, elem.y(), elem.z());
  if (label2 != UNLABELED && label2 != SSL_ELEM) result.insert(label2);
  int label3 = getLabel(elem.x(), elem.y()-1, elem.z());
  if (label3 != UNLABELED && label3 != SSL_ELEM) result.insert(label3);
  int label4 = getLabel(elem.x(), elem.y()+1, elem.z());
  if (label4 != UNLABELED && label4 != SSL_ELEM) result.insert(label4);
  int label5 = getLabel(elem.x(), elem.y(), elem.z()-1);
  if (label5 != UNLABELED && label5 != SSL_ELEM) result.insert(label5);
  int label6 = getLabel(elem.x(), elem.y(), elem.z()+1);
  if (label6 != UNLABELED && label6 != SSL_ELEM) result.insert(label6);
  
  int label7 = getLabel(elem.x()+1, elem.y()+1, elem.z());
  if (label7 != UNLABELED && label7 != SSL_ELEM) result.insert(label7);
  int label8 = getLabel(elem.x()-1, elem.y()-1, elem.z());
  if (label8 != UNLABELED && label8 != SSL_ELEM) result.insert(label8);
  int label9 = getLabel(elem.x()+1, elem.y()-1, elem.z());
  if (label9 != UNLABELED && label9 != SSL_ELEM) result.insert(label9);
  int label10 = getLabel(elem.x()-1, elem.y()+1, elem.z());
  if (label10 != UNLABELED && label10 != SSL_ELEM) result.insert(label10);
  int label11 = getLabel(elem.x(), elem.y()+1, elem.z()+1);
  if (label11 != UNLABELED && label11 != SSL_ELEM) result.insert(label11);
  int label12 = getLabel(elem.x(), elem.y()-1, elem.z()-1);
  if (label12 != UNLABELED && label12 != SSL_ELEM) result.insert(label12);
  int label13 = getLabel(elem.x(), elem.y()+1, elem.z()-1);
  if (label13 != UNLABELED && label13 != SSL_ELEM) result.insert(label13);
  int label14 = getLabel(elem.x(), elem.y()-1, elem.z()+1);
  if (label14 != UNLABELED && label14 != SSL_ELEM) result.insert(label14);
  int label15 = getLabel(elem.x()+1, elem.y(), elem.z()+1);
  if (label15 != UNLABELED && label15 != SSL_ELEM) result.insert(label15);
  int label16 = getLabel(elem.x()-1, elem.y(), elem.z()-1);
  if (label16 != UNLABELED && label16 != SSL_ELEM) result.insert(label16);
  int label17 = getLabel(elem.x()+1, elem.y(), elem.z()-1);
  if (label17 != UNLABELED && label17 != SSL_ELEM) result.insert(label17);
  int label18 = getLabel(elem.x()-1, elem.y(), elem.z()+1);
  if (label18 != UNLABELED && label18 != SSL_ELEM) result.insert(label18);
  
  int label19 = getLabel(elem.x()+1, elem.y()+1, elem.z()+1);
  if (label19 != UNLABELED && label19 != SSL_ELEM) result.insert(label19);
  int label20 = getLabel(elem.x()+1, elem.y()+1, elem.z()-1);
  if (label20 != UNLABELED && label20 != SSL_ELEM) result.insert(label20);
  int label21 = getLabel(elem.x()-1, elem.y()+1, elem.z()+1);
  if (label21 != UNLABELED && label21 != SSL_ELEM) result.insert(label21);
  int label22 = getLabel(elem.x()-1, elem.y()+1, elem.z()-1);
  if (label22 != UNLABELED && label22 != SSL_ELEM) result.insert(label22);
  int label23 = getLabel(elem.x()+1, elem.y()-1, elem.z()+1);
  if (label23 != UNLABELED && label23 != SSL_ELEM) result.insert(label23);
  int label24 = getLabel(elem.x()+1, elem.y()-1, elem.z()-1);
  if (label24 != UNLABELED && label24 != SSL_ELEM) result.insert(label24);
  int label25 = getLabel(elem.x()-1, elem.y()-1, elem.z()+1);
  if (label25 != UNLABELED && label25 != SSL_ELEM) result.insert(label25);
  int label26 = getLabel(elem.x()-1, elem.y()-1, elem.z()-1);
  if (label26 != UNLABELED && label26 != SSL_ELEM) result.insert(label26);
  return result;
}

void Segmentation::SRG3D(vector<struct VoxIndex> seeds, bool sslHeuristicSearch)
{
  sslHeuristicSearchOn = sslHeuristicSearch;
  vector<float> intensitySums;
  vector<float> numIntensities;
  //list<SSLElem> ssl;
  multiset<SSLElem> ssl;
  //reservable_priority_queue<SSLElem> ssl;
  //ssl.reserve(10000000); //2000000 // max: 210000000
  //reservable_priority_queue<SSLElem> ssl2;
  //ssl2.reserve(10000000);
  
  for (int i = 0; i < seeds.size(); ++i) {
    if (!validIndex(seeds.at(i).x, seeds.at(i).y, seeds.at(i).z)) {
      cout << "Seed placed at invalid index!\n";
      exit(1);
    }
    Voxel * seedVoxel = voxels[seeds.at(i).x][seeds.at(i).y][seeds.at(i).z];
    seedVoxel->setLabel(i+1);
    intensitySums.push_back(seedVoxel->getIntensity());
    numIntensities.push_back(1.0f);
    add6NeighborsToSSL_BinarySearchTree(seeds.at(i), ssl, seedVoxel->getIntensity());
    //add26NeighborsToSSL_List(seeds.at(i), ssl, seedVoxel->getIntensity());
    //add26NeighborsToSSL_Heap(seeds.at(i), ssl, seedVoxel->getIntensity());
    //add6NeighborsToSSL_DualHeap(seeds.at(i), ssl, ssl2, seedVoxel->getIntensity());
  }
  //vector<struct VoxIndex>().swap(seeds);
  sslSearchHeuristicInitialized = true;
  
  //printSSL(ssl);
  
  while (!ssl.empty()) {
    //printSSL(ssl);
    //cout << "Size of SSL: " << ssl.size() << endl;
    /*
    if (ssl.size() > 100000 & ssl.size() < 100002) cout << "SSL greater than 100,000...\n";
    if (ssl.size() > 500000 & ssl.size() < 500002) cout << "SSL greater than 500,000...\n";
    if (ssl.size() > 1000000 & ssl.size() < 1000002) cout << "SSL greater than 1,000,000...\n";
    if (ssl.size() > 2000000 & ssl.size() < 2000002) cout << "SSL greater than 2,000,000...\n";
    */
     
    // doubly-linked list
    //SSLElem elem = ssl.front();
    //ssl.pop_front();
    
    // binary search tree
    SSLElem elem(0, 0, 0, 0.0f);
    if (sslSearchHeuristic == ssl.begin()) {
      elem = (*ssl.begin());
      ssl.erase(ssl.begin());
      sslSearchHeuristic = ssl.begin();
    } else {
      elem = (*ssl.begin());
      ssl.erase(ssl.begin());
    }
    
    // heap
    //SSLElem elem = ssl.top();
    //ssl.pop();
    
    // dual heap
    //SSLElem elem = ssl.top();
    //getSSLMinElem(ssl, ssl2, elem);
    
    set<int> neighborLabels = get6NeighborLabels(elem);
    //printSetInfo(neighborLabels);
    if (neighborLabels.size() == 1) {
      int label = *(neighborLabels.begin());
      voxels[elem.x()][elem.y()][elem.z()]->setLabel(label);
      float intensity = voxels[elem.x()][elem.y()][elem.z()]->getIntensity();
      intensitySums[label-1] = intensitySums[label-1] + intensity;
      numIntensities[label-1] = numIntensities[label-1] + 1.0f;
      struct VoxIndex vi = {elem.x(), elem.y(), elem.z()};
      add6NeighborsToSSL_BinarySearchTree(vi, ssl, intensitySums[label-1]/numIntensities[label-1]);
      //add26NeighborsToSSL_List(vi, ssl, intensitySums[label-1]/numIntensities[label-1]);
      //add26NeighborsToSSL_Heap(vi, ssl, intensitySums[label-1]/numIntensities[label-1]);
      //add6NeighborsToSSL_DualHeap(vi, ssl, ssl2, intensitySums[label-1]/numIntensities[label-1]);
    } else {
      
      float intensity = voxels[elem.x()][elem.y()][elem.z()]->getIntensity();
      int curLabel = *(neighborLabels.begin());
      float curDelta = getDelta(intensity, intensitySums[curLabel-1]/numIntensities[curLabel-1]);
      //cout << "Delta: " << curDelta << endl;

      set<int>::iterator it;
      for (it = neighborLabels.begin(); it != neighborLabels.end(); ++it) {
        int itLabel = *it;
        float itDelta = getDelta(intensity, intensitySums[itLabel-1]/numIntensities[itLabel-1]);
        if (itDelta <= curDelta) {
          curLabel = itLabel;
          curDelta = itDelta;
        }
      }
      voxels[elem.x()][elem.y()][elem.z()]->setLabel(curLabel);
      intensitySums[curLabel-1] = intensitySums[curLabel-1] + intensity;
      numIntensities[curLabel-1] = numIntensities[curLabel-1] + 1.0f;
    
    }
    
  }
}

//------------------------------------------------------------------------------

void Segmentation::resetLabels()
{
  for (short x = 0; x < voxels.size(); ++x) {
    for (short y = 0; y < voxels.size(); ++y) {
      for (short z = 0; z < voxels.size(); ++z) {
        if (voxels[x][y][z]) voxels[x][y][z]->setLabel(UNLABELED);
      }
    }
  }
}

vector<struct VoxIndex> Segmentation::findSeeds()
{
  vector<struct VoxIndex> result;
  if (segments.empty()) return result;
  
  for (int segNum = 0; segNum < segments.size(); ++segNum) {
    if (!segments.at(segNum).empty()) {
      //int voxNum = segments[segNum].size() / 2;
      int voxNum = 0;
      short indX = segments[segNum][voxNum][0];
      short indY = segments[segNum][voxNum][1];
      short indZ = segments[segNum][voxNum][2];
      struct VoxIndex voxel = {indX, indY, indZ};
      result.push_back(voxel);
    }
  }
  vector<vector<vector<short> > >().swap(segments);
  vector<vector<vector<short> > >().swap(segmentSurfaces);
  
  resetLabels();
  
  cout << "Seeds:\n";
  for (int i = 0; i < result.size(); ++i) {
    cout << result[i].x << ", " << result[i].y << ", " << result[i].z << endl;
  }
  
  return result;
}

