// Voxel.cpp

#include <iostream>
#include "Voxel.h"
#include "Globals.h"
#include <glut/glut.h>

using std::cout;
using std::endl;

float Voxel::alpha = 1.0;
float Voxel::dim = .1;
bool Voxel::pseudoColor = false;
bool Voxel::wireframe = false;
bool Voxel::point = true;
const float Voxel::MAX_VAL = 2500.0; // 2500.0 for REB medical volume data // 65535.0

Voxel::Voxel() {}

Voxel::Voxel(float voxVal)
{
  vox_data.intensity = floatToIntensity(voxVal);
  vox_data.label = UNLABELED;
  //cout << "Value of label: " << vox_data.label << endl;
}

Voxel::Voxel(unsigned short voxVal)
{
  vox_data.intensity = voxVal;
  vox_data.label = UNLABELED;
}

unsigned short Voxel::floatToIntensity(float fVal)
{
  return (unsigned short)(MAX_VAL * fVal);
}

float Voxel::intensityToFloat()
{
  float tempVal = ((float)vox_data.intensity) / MAX_VAL;
  if (tempVal > 1.0) tempVal = 1.0;
  else if (tempVal < 0.0) tempVal = 0.0;
  return tempVal;
}

Voxel::~Voxel() {}

void Voxel::draw(short& gridX, short& gridY, short& gridZ)
{
  float x = (float)gridX * dim; // startX
  float y = (float)gridY * dim; // startY
  float z = (float)gridZ * -dim; // startZ
  
  float fVal = intensityToFloat();
  if (pseudoColor) {
    if (vox_data.label == 1) glColor4f(1.0f, 0.0f, 0.0f, alpha);
    else if (vox_data.label == 2) glColor4f(0.0f, 1.0f, 0.0f, alpha);
    else if (vox_data.label == 3) glColor4f(0.0f, 0.0f, 1.0f, alpha);
    else if (vox_data.label == 4) glColor4f(1.0f, 1.0f, 0.0f, alpha);
    else if (vox_data.label == 5) glColor4f(1.0f, 0.0f, 1.0f, alpha);
    else if (vox_data.label == 6) glColor4f(0.0f, 1.0f, 1.0f, alpha);
    else if (vox_data.label == 7) glColor4f(1.0f, 0.5f, 0.5f, alpha);
    else if (vox_data.label == 8) glColor4f(0.5f, 1.0f, 0.5f, alpha);
    else if (vox_data.label == 9) glColor4f(0.5f, 0.5f, 1.0f, alpha);
    else if (vox_data.label == 10) glColor4f(1.0f, 1.0f, 0.5f, alpha);
    else if (vox_data.label == 11) glColor4f(1.0f, 0.5f, 1.0f, alpha);
    else if (vox_data.label == 12) glColor4f(0.5f, 1.0f, 1.0f, alpha);
    else if (vox_data.label == 13) glColor4f(0.3f, 1.0f, 0.0f, alpha);
    else if (vox_data.label == 14) glColor4f(0.4f, 0.0f, 1.0f, alpha);
    else if (vox_data.label == 15) glColor4f(0.5f, 1.0f, 0.0f, alpha);
    else if (vox_data.label == 16) glColor4f(0.6f, 0.0f, 1.0f, alpha);
    else if (vox_data.label == 17) glColor4f(0.7f, 1.0f, 1.0f, alpha);
    else if (vox_data.label == 18) glColor4f(0.8f, 0.5f, 0.5f, alpha);
    else if (vox_data.label == 19) glColor4f(0.9f, 1.0f, 0.5f, alpha);
    else if (vox_data.label == 20) glColor4f(0.2f, 0.5f, 1.0f, alpha);
    else if (vox_data.label == 21) glColor4f(0.1f, 1.0f, 0.5f, alpha);
    else if (vox_data.label == 22) glColor4f(1.0f, 0.1f, 1.0f, alpha);
    else if (vox_data.label == 23) glColor4f(0.5f, 0.2f, 1.0f, alpha);
    else if (vox_data.label == 24) glColor4f(0.0f, 0.3f, 0.0f, alpha);
    else if (vox_data.label == 25) glColor4f(0.0f, 0.4f, 1.0f, alpha);
    else if (vox_data.label == 26) glColor4f(1.0f, 0.5f, 0.0f, alpha);
    else if (vox_data.label == 27) glColor4f(1.0f, 0.6f, 1.0f, alpha);
    else if (vox_data.label == 28) glColor4f(0.0f, 0.7f, 1.0f, alpha);
    else if (vox_data.label == 29) glColor4f(1.0f, 0.8f, 0.5f, alpha);
    else if (vox_data.label == 30) glColor4f(0.5f, 0.9f, 0.5f, alpha);
    else if (vox_data.label == 31) glColor4f(0.5f, 0.5f, 0.1f, alpha);
    else if (vox_data.label == 32) glColor4f(1.0f, 1.0f, 0.2f, alpha);
    else if (vox_data.label == 33) glColor4f(1.0f, 0.5f, 0.3f, alpha);
    else if (vox_data.label == 34) glColor4f(0.5f, 1.0f, 0.4f, alpha);
    else if (vox_data.label == 35) glColor4f(0.3f, 1.0f, 0.5f, alpha);
    else if (vox_data.label == 36) glColor4f(0.4f, 0.0f, 0.6f, alpha);
    else if (vox_data.label == 37) glColor4f(0.5f, 1.0f, 0.7f, alpha);
    else if (vox_data.label == 38) glColor4f(0.6f, 0.0f, 0.8f, alpha);
    else if (vox_data.label == 39) glColor4f(0.7f, 1.0f, 0.9f, alpha);
    else if (vox_data.label == 40) glColor4f(0.3f, 0.2f, 0.1f, alpha);
    else if (vox_data.label == 41) glColor4f(0.1f, 0.2f, 0.3f, alpha);
    else if (vox_data.label == 42) glColor4f(0.3f, 0.4f, 0.3f, alpha);
    else if (vox_data.label == 43) glColor4f(0.1f, 0.1f, 0.4f, alpha);
    else if (vox_data.label == 44) glColor4f(0.9f, 0.2f, 0.2f, alpha);
    else if (vox_data.label == 45) glColor4f(0.9f, 0.9f, 0.7f, alpha);
    else if (vox_data.label == 46) glColor4f(0.9f, 0.0f, 0.1f, alpha);
    else if (vox_data.label == 47) glColor4f(0.8f, 0.0f, 0.2f, alpha);
    else if (vox_data.label == 48) glColor4f(0.7f, 0.0f, 0.3f, alpha);
    else if (vox_data.label == 49) glColor4f(0.6f, 0.0f, 0.4f, alpha);
    else if (vox_data.label == 50) glColor4f(0.5f, 0.0f, 0.5f, alpha);
    else if (vox_data.label == 51) glColor4f(0.4f, 0.0f, 0.6f, alpha);
    else if (vox_data.label == 52) glColor4f(0.3f, 0.0f, 0.7f, alpha);
    else if (vox_data.label == 53) glColor4f(0.2f, 0.0f, 0.8f, alpha);
    else if (vox_data.label == 54) glColor4f(0.1f, 0.1f, 0.9f, alpha);
    else if (vox_data.label == 55) glColor4f(1.0f, 0.2f, 0.25f, alpha);
    else if (vox_data.label == 56) glColor4f(0.5f, 1.3f, 0.8f, alpha);
    else if (vox_data.label == 57) glColor4f(0.3f, 1.4f, 0.7f, alpha);
    else if (vox_data.label == 58) glColor4f(0.4f, 0.5f, 0.6f, alpha);
    else if (vox_data.label == 59) glColor4f(0.1f, 0.6f, 0.5f, alpha);
    else if (vox_data.label == 60) glColor4f(1.0f, 0.7f, 0.4f, alpha);
    else if (vox_data.label == 61) glColor4f(0.5f, 1.8f, 0.3f, alpha);
    else if (vox_data.label == 62) glColor4f(0.3f, 1.9f, 0.2f, alpha);
    else if (vox_data.label == 63) glColor4f(0.4f, 1.0f, 0.1f, alpha);
    else if (vox_data.label == UNLABELED) glColor4f(fVal, fVal, fVal, alpha);
    else if (vox_data.label == NOISE) glColor4f(0.2f, 0.05f, 0.05f, alpha);
    else glColor4f(fVal, fVal, fVal, alpha);
  } else {
    glColor4f(fVal, fVal, fVal, alpha);
  }
  if (alpha != 1.0) glDepthMask(GL_FALSE);
  else glDepthMask(GL_TRUE);
  
  if (!point) {
    
    GLfloat vertices[] =
    {x,y,z,  x+dim,y,z,  x+dim,y+dim,z,  x,y+dim,z, // FRONT
      x,y,z-dim,  x,y+dim,z-dim,  x+dim,y+dim,z-dim,  x+dim,y,z-dim, // BACK
      x+dim,y,z,  x+dim,y,z-dim,  x+dim,y+dim,z-dim,  x+dim,y+dim,z, // RIGHT
      x,y,z,  x,y+dim,z,  x,y+dim,z-dim,  x,y,z-dim, // LEFT
      x,y,z,  x,y,z-dim,  x+dim,y,z-dim,  x+dim,y,z, // BOTTOM
    x,y+dim,z,  x+dim,y+dim,z,  x+dim,y+dim,z-dim,  x,y+dim,z-dim}; // TOP
    
    GLfloat normals[] =
    {0,0,1,  0,0,1,  0,0,1,  0,0,1,
      0,0,-1,  0,0,-1,  0,0,-1,  0,0,-1,
      1,0,0,  1,0,0,  1,0,0,  1,0,0,
      -1,0,0,  -1,0,0,  -1,0,0,  -1,0,0,
      0,-1,0,  0,-1,0,  0,-1,0,  0,-1,0,
    0,1,0,  0,1,0,  0,1,0,  0,1,0};
    
    glNormalPointer(GL_FLOAT, 0, normals);
    glVertexPointer(3, GL_FLOAT, 0, vertices);
    GLubyte boxIndices[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23};
    
    glDrawElements(GL_QUADS, 24, GL_UNSIGNED_BYTE, boxIndices);
    
    if (wireframe) {
      glColor4f(0.0, 0.0, 0.0, 1.0);
      glPolygonOffset(0.1, 0.2);
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      glLineWidth(0.1); // 0.01 (far) // 0.2 (close)
      glDrawElements(GL_QUADS, 24, GL_UNSIGNED_BYTE, boxIndices);
      glPolygonOffset(0.0, 0.0);
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }
  }
  
  if (point) {
    GLfloat vertices[] = {x+(dim/2.0f), y+(dim/2.0f), z+(dim/2.0f)};
    GLfloat normals[] = {0,0,1};
    
    glNormalPointer(GL_FLOAT, 0, normals);
    glVertexPointer(3, GL_FLOAT, 0, vertices);
    GLubyte pointIndices[] = {0};
  
    glDrawElements(GL_POINTS, 1, GL_UNSIGNED_BYTE, pointIndices);
  }
  
}
